"""File generated by TLObjects' generator. All changes will be ERASED"""
from ...tl.tlobject import TLObject


class CdnFile(TLObject):
    constructor_id = 0xa99fca4f
    subclass_of_id = 0xf5ccf928

    def __init__(self, bytes):
        """
        :param bytes: Telegram type: "bytes".

        Constructor for upload.CdnFile: Instance of either CdnFileReuploadNeeded, CdnFile.
        """
        super().__init__()

        self.bytes = bytes

    def to_dict(self):
        return {
            'bytes': self.bytes,
        }

    def on_send(self, writer):
        writer.write_int(CdnFile.constructor_id, signed=False)
        writer.tgwrite_bytes(self.bytes)

    @staticmethod
    def empty():
        return CdnFile(None)

    def on_response(self, reader):
        self.bytes = reader.tgread_bytes()

    def __repr__(self):
        return 'upload.cdnFile#a99fca4f bytes:bytes = upload.CdnFile'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class CdnFileReuploadNeeded(TLObject):
    constructor_id = 0xeea8e46e
    subclass_of_id = 0xf5ccf928

    def __init__(self, request_token):
        """
        :param request_token: Telegram type: "bytes".

        Constructor for upload.CdnFile: Instance of either CdnFileReuploadNeeded, CdnFile.
        """
        super().__init__()

        self.request_token = request_token

    def to_dict(self):
        return {
            'request_token': self.request_token,
        }

    def on_send(self, writer):
        writer.write_int(CdnFileReuploadNeeded.constructor_id, signed=False)
        writer.tgwrite_bytes(self.request_token)

    @staticmethod
    def empty():
        return CdnFileReuploadNeeded(None)

    def on_response(self, reader):
        self.request_token = reader.tgread_bytes()

    def __repr__(self):
        return 'upload.cdnFileReuploadNeeded#eea8e46e request_token:bytes = upload.CdnFile'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class File(TLObject):
    constructor_id = 0x96a18d5
    subclass_of_id = 0x6c9bd728

    def __init__(self, type, mtime, bytes):
        """
        :param type: Telegram type: "storage.FileType".
        :param mtime: Telegram type: "int".
        :param bytes: Telegram type: "bytes".

        Constructor for upload.File: Instance of either File, FileCdnRedirect.
        """
        super().__init__()

        self.type = type
        self.mtime = mtime
        self.bytes = bytes

    def to_dict(self):
        return {
            'type': None if self.type is None else self.type.to_dict(),
            'mtime': self.mtime,
            'bytes': self.bytes,
        }

    def on_send(self, writer):
        writer.write_int(File.constructor_id, signed=False)
        self.type.on_send(writer)
        writer.write_int(self.mtime)
        writer.tgwrite_bytes(self.bytes)

    @staticmethod
    def empty():
        return File(None, None, None)

    def on_response(self, reader):
        self.type = reader.tgread_object()
        self.mtime = reader.read_int()
        self.bytes = reader.tgread_bytes()

    def __repr__(self):
        return 'upload.file#096a18d5 type:storage.FileType mtime:int bytes:bytes = upload.File'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class FileCdnRedirect(TLObject):
    constructor_id = 0xea52fe5a
    subclass_of_id = 0x6c9bd728

    def __init__(self, dc_id, file_token, encryption_key, encryption_iv, cdn_file_hashes):
        """
        :param dc_id: Telegram type: "int".
        :param file_token: Telegram type: "bytes".
        :param encryption_key: Telegram type: "bytes".
        :param encryption_iv: Telegram type: "bytes".
        :param cdn_file_hashes: Telegram type: "CdnFileHash". Must be a list.

        Constructor for upload.File: Instance of either File, FileCdnRedirect.
        """
        super().__init__()

        self.dc_id = dc_id
        self.file_token = file_token
        self.encryption_key = encryption_key
        self.encryption_iv = encryption_iv
        self.cdn_file_hashes = cdn_file_hashes

    def to_dict(self):
        return {
            'dc_id': self.dc_id,
            'file_token': self.file_token,
            'encryption_key': self.encryption_key,
            'encryption_iv': self.encryption_iv,
            'cdn_file_hashes': [] if self.cdn_file_hashes is None else [None if x is None else x.to_dict() for x in self.cdn_file_hashes],
        }

    def on_send(self, writer):
        writer.write_int(FileCdnRedirect.constructor_id, signed=False)
        writer.write_int(self.dc_id)
        writer.tgwrite_bytes(self.file_token)
        writer.tgwrite_bytes(self.encryption_key)
        writer.tgwrite_bytes(self.encryption_iv)
        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.cdn_file_hashes))
        for _x in self.cdn_file_hashes:
            _x.on_send(writer)

    @staticmethod
    def empty():
        return FileCdnRedirect(None, None, None, None, None)

    def on_response(self, reader):
        self.dc_id = reader.read_int()
        self.file_token = reader.tgread_bytes()
        self.encryption_key = reader.tgread_bytes()
        self.encryption_iv = reader.tgread_bytes()
        reader.read_int()
        self.cdn_file_hashes = []
        _len = reader.read_int()
        for _ in range(_len):
            _x = reader.tgread_object()
            self.cdn_file_hashes.append(_x)

    def __repr__(self):
        return 'upload.fileCdnRedirect#ea52fe5a dc_id:int file_token:bytes encryption_key:bytes encryption_iv:bytes cdn_file_hashes:Vector<CdnFileHash> = upload.File'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class WebFile(TLObject):
    constructor_id = 0x21e753bc
    subclass_of_id = 0x68f17f51

    def __init__(self, size, mime_type, file_type, mtime, bytes):
        """
        :param size: Telegram type: "int".
        :param mime_type: Telegram type: "string".
        :param file_type: Telegram type: "storage.FileType".
        :param mtime: Telegram type: "int".
        :param bytes: Telegram type: "bytes".

        Constructor for upload.WebFile: Instance of WebFile.
        """
        super().__init__()

        self.size = size
        self.mime_type = mime_type
        self.file_type = file_type
        self.mtime = mtime
        self.bytes = bytes

    def to_dict(self):
        return {
            'size': self.size,
            'mime_type': self.mime_type,
            'file_type': None if self.file_type is None else self.file_type.to_dict(),
            'mtime': self.mtime,
            'bytes': self.bytes,
        }

    def on_send(self, writer):
        writer.write_int(WebFile.constructor_id, signed=False)
        writer.write_int(self.size)
        writer.tgwrite_string(self.mime_type)
        self.file_type.on_send(writer)
        writer.write_int(self.mtime)
        writer.tgwrite_bytes(self.bytes)

    @staticmethod
    def empty():
        return WebFile(None, None, None, None, None)

    def on_response(self, reader):
        self.size = reader.read_int()
        self.mime_type = reader.tgread_string()
        self.file_type = reader.tgread_object()
        self.mtime = reader.read_int()
        self.bytes = reader.tgread_bytes()

    def __repr__(self):
        return 'upload.webFile#21e753bc size:int mime_type:string file_type:storage.FileType mtime:int bytes:bytes = upload.WebFile'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)




"""File generated by TLObjects' generator. All changes will be ERASED"""
from ...tl.tlobject import TLObject


class GetDifferenceRequest(TLObject):
    constructor_id = 0xb2e4d7d
    subclass_of_id = 0x52662d55

    def __init__(self, from_version):
        """
        :param from_version: Telegram type: "int".

        :returns LangPackDifference: Instance of LangPackDifference.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.from_version = from_version

    def to_dict(self):
        return {
            'from_version': self.from_version,
        }

    def on_send(self, writer):
        writer.write_int(GetDifferenceRequest.constructor_id, signed=False)
        writer.write_int(self.from_version)

    @staticmethod
    def empty():
        return GetDifferenceRequest(None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'langpack.getDifference#0b2e4d7d from_version:int = LangPackDifference'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class GetLangPackRequest(TLObject):
    constructor_id = 0x9ab5c58e
    subclass_of_id = 0x52662d55

    def __init__(self, lang_code):
        """
        :param lang_code: Telegram type: "string".

        :returns LangPackDifference: Instance of LangPackDifference.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.lang_code = lang_code

    def to_dict(self):
        return {
            'lang_code': self.lang_code,
        }

    def on_send(self, writer):
        writer.write_int(GetLangPackRequest.constructor_id, signed=False)
        writer.tgwrite_string(self.lang_code)

    @staticmethod
    def empty():
        return GetLangPackRequest(None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'langpack.getLangPack#9ab5c58e lang_code:string = LangPackDifference'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class GetLanguagesRequest(TLObject):
    constructor_id = 0x800fd57d
    subclass_of_id = 0x280912c9

    def __init__(self):
        super().__init__()
        self.result = None
        self.content_related = True

    @staticmethod
    def to_dict():
        return {}

    def on_send(self, writer):
        writer.write_int(GetLanguagesRequest.constructor_id, signed=False)

    @staticmethod
    def empty():
        return GetLanguagesRequest()

    def on_response(self, reader):
        self.result = reader.tgread_vector()

    def __repr__(self):
        return 'langpack.getLanguages#800fd57d = Vector<LangPackLanguage>'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class GetStringsRequest(TLObject):
    constructor_id = 0x2e1ee318
    subclass_of_id = 0xc7b7353d

    def __init__(self, lang_code, keys):
        """
        :param lang_code: Telegram type: "string".
        :param keys: Telegram type: "string". Must be a list.

        :returns Vector<LangPackString>: This type has no constructors.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.lang_code = lang_code
        self.keys = keys

    def to_dict(self):
        return {
            'lang_code': self.lang_code,
            'keys': [] if self.keys is None else self.keys[:],
        }

    def on_send(self, writer):
        writer.write_int(GetStringsRequest.constructor_id, signed=False)
        writer.tgwrite_string(self.lang_code)
        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.keys))
        for _x in self.keys:
            writer.tgwrite_string(_x)

    @staticmethod
    def empty():
        return GetStringsRequest(None, None)

    def on_response(self, reader):
        self.result = reader.tgread_vector()

    def __repr__(self):
        return 'langpack.getStrings#2e1ee318 lang_code:string keys:Vector<string> = Vector<LangPackString>'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)




"""File generated by TLObjects' generator. All changes will be ERASED"""
from ...tl.tlobject import TLObject


class AddStickerToSetRequest(TLObject):
    constructor_id = 0x8653febe
    subclass_of_id = 0x9b704a5a

    def __init__(self, stickerset, sticker):
        """
        :param stickerset: Telegram type: "InputStickerSet".
        :param sticker: Telegram type: "InputStickerSetItem".

        :returns messages.StickerSet: Instance of StickerSet.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.stickerset = stickerset
        self.sticker = sticker

    def to_dict(self):
        return {
            'stickerset': None if self.stickerset is None else self.stickerset.to_dict(),
            'sticker': None if self.sticker is None else self.sticker.to_dict(),
        }

    def on_send(self, writer):
        writer.write_int(AddStickerToSetRequest.constructor_id, signed=False)
        self.stickerset.on_send(writer)
        self.sticker.on_send(writer)

    @staticmethod
    def empty():
        return AddStickerToSetRequest(None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'stickers.addStickerToSet#8653febe stickerset:InputStickerSet sticker:InputStickerSetItem = messages.StickerSet'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class ChangeStickerPositionRequest(TLObject):
    constructor_id = 0xffb6d4ca
    subclass_of_id = 0x9b704a5a

    def __init__(self, sticker, position):
        """
        :param sticker: Telegram type: "InputDocument".
        :param position: Telegram type: "int".

        :returns messages.StickerSet: Instance of StickerSet.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.sticker = sticker
        self.position = position

    def to_dict(self):
        return {
            'sticker': None if self.sticker is None else self.sticker.to_dict(),
            'position': self.position,
        }

    def on_send(self, writer):
        writer.write_int(ChangeStickerPositionRequest.constructor_id, signed=False)
        self.sticker.on_send(writer)
        writer.write_int(self.position)

    @staticmethod
    def empty():
        return ChangeStickerPositionRequest(None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'stickers.changeStickerPosition#ffb6d4ca sticker:InputDocument position:int = messages.StickerSet'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)

from ...utils import get_input_user


class CreateStickerSetRequest(TLObject):
    constructor_id = 0x9bd86e6a
    subclass_of_id = 0x9b704a5a

    def __init__(self, user_id, title, short_name, stickers, masks=None):
        """
        :param masks: Telegram type: "true".
        :param user_id: Telegram type: "InputUser".
        :param title: Telegram type: "string".
        :param short_name: Telegram type: "string".
        :param stickers: Telegram type: "InputStickerSetItem". Must be a list.

        :returns messages.StickerSet: Instance of StickerSet.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.masks = masks
        self.user_id = get_input_user(user_id)
        self.title = title
        self.short_name = short_name
        self.stickers = stickers

    def to_dict(self):
        return {
            'masks': self.masks,
            'user_id': None if self.user_id is None else self.user_id.to_dict(),
            'title': self.title,
            'short_name': self.short_name,
            'stickers': [] if self.stickers is None else [None if x is None else x.to_dict() for x in self.stickers],
        }

    def on_send(self, writer):
        writer.write_int(CreateStickerSetRequest.constructor_id, signed=False)
        flags = 0
        flags |= (1 << 0) if self.masks else 0
        writer.write_int(flags)

        self.user_id.on_send(writer)
        writer.tgwrite_string(self.title)
        writer.tgwrite_string(self.short_name)
        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.stickers))
        for _x in self.stickers:
            _x.on_send(writer)

    @staticmethod
    def empty():
        return CreateStickerSetRequest(None, None, None, None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'stickers.createStickerSet#9bd86e6a flags:# masks:flags.0?true user_id:InputUser title:string short_name:string stickers:Vector<InputStickerSetItem> = messages.StickerSet'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class RemoveStickerFromSetRequest(TLObject):
    constructor_id = 0xf7760f51
    subclass_of_id = 0x9b704a5a

    def __init__(self, sticker):
        """
        :param sticker: Telegram type: "InputDocument".

        :returns messages.StickerSet: Instance of StickerSet.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.sticker = sticker

    def to_dict(self):
        return {
            'sticker': None if self.sticker is None else self.sticker.to_dict(),
        }

    def on_send(self, writer):
        writer.write_int(RemoveStickerFromSetRequest.constructor_id, signed=False)
        self.sticker.on_send(writer)

    @staticmethod
    def empty():
        return RemoveStickerFromSetRequest(None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'stickers.removeStickerFromSet#f7760f51 sticker:InputDocument = messages.StickerSet'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)




from io import BufferedWriter, BytesIO, DEFAULT_BUFFER_SIZE
from struct import pack


class BinaryWriter:
    """
    Small utility class to write binary data.
    Also creates a "Memory Stream" if necessary
    """

    def __init__(self, stream=None, known_length=None):
        if not stream:
            stream = BytesIO()

        if known_length is None:
            # On some systems, DEFAULT_BUFFER_SIZE defaults to 8192
            # That's over 16 times as big as necessary for most messages
            known_length = max(DEFAULT_BUFFER_SIZE, 1024)

        self.writer = BufferedWriter(stream, buffer_size=known_length)
        self.written_count = 0

    # region Writing

    # "All numbers are written as little endian."
    # https://core.telegram.org/mtproto
    def write_byte(self, value):
        """Writes a single byte value"""
        self.writer.write(pack('B', value))
        self.written_count += 1

    def write_int(self, value, signed=True):
        """Writes an integer value (4 bytes), optionally signed"""
        self.writer.write(
            int.to_bytes(
                value, length=4, byteorder='little', signed=signed))
        self.written_count += 4

    def write_long(self, value, signed=True):
        """Writes a long integer value (8 bytes), optionally signed"""
        self.writer.write(
            int.to_bytes(
                value, length=8, byteorder='little', signed=signed))
        self.written_count += 8

    def write_float(self, value):
        """Writes a floating point value (4 bytes)"""
        self.writer.write(pack('<f', value))
        self.written_count += 4

    def write_double(self, value):
        """Writes a floating point value (8 bytes)"""
        self.writer.write(pack('<d', value))
        self.written_count += 8

    def write_large_int(self, value, bits, signed=True):
        """Writes a n-bits long integer value"""
        self.writer.write(
            int.to_bytes(
                value, length=bits // 8, byteorder='little', signed=signed))
        self.written_count += bits // 8

    def write(self, data):
        """Writes the given bytes array"""
        self.writer.write(data)
        self.written_count += len(data)

    # endregion

    # region Telegram custom writing

    def tgwrite_bytes(self, data):
        """Write bytes by using Telegram guidelines"""
        if len(data) < 254:
            padding = (len(data) + 1) % 4
            if padding != 0:
                padding = 4 - padding

            self.write(bytes([len(data)]))
            self.write(data)

        else:
            padding = len(data) % 4
            if padding != 0:
                padding = 4 - padding

            self.write(bytes([254]))
            self.write(bytes([len(data) % 256]))
            self.write(bytes([(len(data) >> 8) % 256]))
            self.write(bytes([(len(data) >> 16) % 256]))
            self.write(data)

        self.write(bytes(padding))

    def tgwrite_string(self, string):
        """Write a string by using Telegram guidelines"""
        self.tgwrite_bytes(string.encode('utf-8'))

    def tgwrite_bool(self, boolean):
        """Write a boolean value by using Telegram guidelines"""
        #                     boolTrue                boolFalse
        self.write_int(0x997275b5 if boolean else 0xbc799737, signed=False)

    def tgwrite_date(self, datetime):
        """Converts a Python datetime object into Unix time
           (used by Telegram) and writes it
        """
        value = 0 if datetime is None else int(datetime.timestamp())
        self.write_int(value)

    def tgwrite_object(self, tlobject):
        """Writes a Telegram object"""
        tlobject.on_send(self)

    def tgwrite_vector(self, vector):
        """Writes a vector of Telegram objects"""
        self.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        self.write_int(len(vector))
        for item in vector:
            self.tgwrite_object(item)

    # endregion

    def flush(self):
        """Flush the current stream to "update" changes"""
        self.writer.flush()

    def close(self):
        """Close the current stream"""
        self.writer.close()

    def get_bytes(self, flush=True):
        """Get the current bytes array content from the buffer,
           optionally flushing first
        """
        if flush:
            self.writer.flush()
        return self.writer.raw.getvalue()

    def get_written_bytes_count(self):
        """Gets the count of bytes written in the buffer.
           This may NOT be equal to the stream length if one
           was provided when initializing the writer
        """
        return self.written_count

    # with block
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()



"""File generated by TLObjects' generator. All changes will be ERASED"""
from ...tl.tlobject import TLObject


class AcceptCallRequest(TLObject):
    constructor_id = 0x3bd2b4a0
    subclass_of_id = 0xd48afe4f

    def __init__(self, peer, g_b, protocol):
        """
        :param peer: Telegram type: "InputPhoneCall".
        :param g_b: Telegram type: "bytes".
        :param protocol: Telegram type: "PhoneCallProtocol".

        :returns phone.PhoneCall: Instance of PhoneCall.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.peer = peer
        self.g_b = g_b
        self.protocol = protocol

    def to_dict(self):
        return {
            'peer': None if self.peer is None else self.peer.to_dict(),
            'g_b': self.g_b,
            'protocol': None if self.protocol is None else self.protocol.to_dict(),
        }

    def on_send(self, writer):
        writer.write_int(AcceptCallRequest.constructor_id, signed=False)
        self.peer.on_send(writer)
        writer.tgwrite_bytes(self.g_b)
        self.protocol.on_send(writer)

    @staticmethod
    def empty():
        return AcceptCallRequest(None, None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'phone.acceptCall#3bd2b4a0 peer:InputPhoneCall g_b:bytes protocol:PhoneCallProtocol = phone.PhoneCall'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class ConfirmCallRequest(TLObject):
    constructor_id = 0x2efe1722
    subclass_of_id = 0xd48afe4f

    def __init__(self, peer, g_a, key_fingerprint, protocol):
        """
        :param peer: Telegram type: "InputPhoneCall".
        :param g_a: Telegram type: "bytes".
        :param key_fingerprint: Telegram type: "long".
        :param protocol: Telegram type: "PhoneCallProtocol".

        :returns phone.PhoneCall: Instance of PhoneCall.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.peer = peer
        self.g_a = g_a
        self.key_fingerprint = key_fingerprint
        self.protocol = protocol

    def to_dict(self):
        return {
            'peer': None if self.peer is None else self.peer.to_dict(),
            'g_a': self.g_a,
            'key_fingerprint': self.key_fingerprint,
            'protocol': None if self.protocol is None else self.protocol.to_dict(),
        }

    def on_send(self, writer):
        writer.write_int(ConfirmCallRequest.constructor_id, signed=False)
        self.peer.on_send(writer)
        writer.tgwrite_bytes(self.g_a)
        writer.write_long(self.key_fingerprint)
        self.protocol.on_send(writer)

    @staticmethod
    def empty():
        return ConfirmCallRequest(None, None, None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'phone.confirmCall#2efe1722 peer:InputPhoneCall g_a:bytes key_fingerprint:long protocol:PhoneCallProtocol = phone.PhoneCall'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class DiscardCallRequest(TLObject):
    constructor_id = 0x78d413a6
    subclass_of_id = 0x8af52aac

    def __init__(self, peer, duration, reason, connection_id):
        """
        :param peer: Telegram type: "InputPhoneCall".
        :param duration: Telegram type: "int".
        :param reason: Telegram type: "PhoneCallDiscardReason".
        :param connection_id: Telegram type: "long".

        :returns Updates: Instance of either UpdatesTooLong, UpdateShortMessage, UpdateShortChatMessage, UpdateShort, UpdatesCombined, Updates, UpdateShortSentMessage.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.peer = peer
        self.duration = duration
        self.reason = reason
        self.connection_id = connection_id

    def to_dict(self):
        return {
            'peer': None if self.peer is None else self.peer.to_dict(),
            'duration': self.duration,
            'reason': None if self.reason is None else self.reason.to_dict(),
            'connection_id': self.connection_id,
        }

    def on_send(self, writer):
        writer.write_int(DiscardCallRequest.constructor_id, signed=False)
        self.peer.on_send(writer)
        writer.write_int(self.duration)
        self.reason.on_send(writer)
        writer.write_long(self.connection_id)

    @staticmethod
    def empty():
        return DiscardCallRequest(None, None, None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'phone.discardCall#78d413a6 peer:InputPhoneCall duration:int reason:PhoneCallDiscardReason connection_id:long = Updates'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class GetCallConfigRequest(TLObject):
    constructor_id = 0x55451fa9
    subclass_of_id = 0xad0352e8

    def __init__(self):
        super().__init__()
        self.result = None
        self.content_related = True

    @staticmethod
    def to_dict():
        return {}

    def on_send(self, writer):
        writer.write_int(GetCallConfigRequest.constructor_id, signed=False)

    @staticmethod
    def empty():
        return GetCallConfigRequest()

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'phone.getCallConfig#55451fa9 = DataJSON'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class ReceivedCallRequest(TLObject):
    constructor_id = 0x17d54f61
    subclass_of_id = 0xf5b399ac

    def __init__(self, peer):
        """
        :param peer: Telegram type: "InputPhoneCall".

        :returns Bool: This type has no constructors.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.peer = peer

    def to_dict(self):
        return {
            'peer': None if self.peer is None else self.peer.to_dict(),
        }

    def on_send(self, writer):
        writer.write_int(ReceivedCallRequest.constructor_id, signed=False)
        self.peer.on_send(writer)

    @staticmethod
    def empty():
        return ReceivedCallRequest(None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'phone.receivedCall#17d54f61 peer:InputPhoneCall = Bool'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)

from ...utils import get_input_user
import os


class RequestCallRequest(TLObject):
    constructor_id = 0x5b95b3d4
    subclass_of_id = 0xd48afe4f

    def __init__(self, user_id, g_a_hash, protocol, random_id=None):
        """
        :param user_id: Telegram type: "InputUser".
        :param random_id: Telegram type: "int".
        :param g_a_hash: Telegram type: "bytes".
        :param protocol: Telegram type: "PhoneCallProtocol".

        :returns phone.PhoneCall: Instance of PhoneCall.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.user_id = get_input_user(user_id)
        self.random_id = random_id if random_id is not None else int.from_bytes(os.urandom(4), signed=True, byteorder='little')
        self.g_a_hash = g_a_hash
        self.protocol = protocol

    def to_dict(self):
        return {
            'user_id': None if self.user_id is None else self.user_id.to_dict(),
            'random_id': self.random_id,
            'g_a_hash': self.g_a_hash,
            'protocol': None if self.protocol is None else self.protocol.to_dict(),
        }

    def on_send(self, writer):
        writer.write_int(RequestCallRequest.constructor_id, signed=False)
        self.user_id.on_send(writer)
        writer.write_int(self.random_id)
        writer.tgwrite_bytes(self.g_a_hash)
        self.protocol.on_send(writer)

    @staticmethod
    def empty():
        return RequestCallRequest(None, None, None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'phone.requestCall#5b95b3d4 user_id:InputUser random_id:int g_a_hash:bytes protocol:PhoneCallProtocol = phone.PhoneCall'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class SaveCallDebugRequest(TLObject):
    constructor_id = 0x277add7e
    subclass_of_id = 0xf5b399ac

    def __init__(self, peer, debug):
        """
        :param peer: Telegram type: "InputPhoneCall".
        :param debug: Telegram type: "DataJSON".

        :returns Bool: This type has no constructors.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.peer = peer
        self.debug = debug

    def to_dict(self):
        return {
            'peer': None if self.peer is None else self.peer.to_dict(),
            'debug': None if self.debug is None else self.debug.to_dict(),
        }

    def on_send(self, writer):
        writer.write_int(SaveCallDebugRequest.constructor_id, signed=False)
        self.peer.on_send(writer)
        self.debug.on_send(writer)

    @staticmethod
    def empty():
        return SaveCallDebugRequest(None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'phone.saveCallDebug#277add7e peer:InputPhoneCall debug:DataJSON = Bool'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class SetCallRatingRequest(TLObject):
    constructor_id = 0x1c536a34
    subclass_of_id = 0x8af52aac

    def __init__(self, peer, rating, comment):
        """
        :param peer: Telegram type: "InputPhoneCall".
        :param rating: Telegram type: "int".
        :param comment: Telegram type: "string".

        :returns Updates: Instance of either UpdatesTooLong, UpdateShortMessage, UpdateShortChatMessage, UpdateShort, UpdatesCombined, Updates, UpdateShortSentMessage.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.peer = peer
        self.rating = rating
        self.comment = comment

    def to_dict(self):
        return {
            'peer': None if self.peer is None else self.peer.to_dict(),
            'rating': self.rating,
            'comment': self.comment,
        }

    def on_send(self, writer):
        writer.write_int(SetCallRatingRequest.constructor_id, signed=False)
        self.peer.on_send(writer)
        writer.write_int(self.rating)
        writer.tgwrite_string(self.comment)

    @staticmethod
    def empty():
        return SetCallRatingRequest(None, None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'phone.setCallRating#1c536a34 peer:InputPhoneCall rating:int comment:string = Updates'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)




import os
from hashlib import sha1
try:
    import rsa
    import rsa.core
except ImportError:
    rsa = None
    raise ImportError('Missing module "rsa", please install via pip.')

from ..extensions import BinaryWriter


# {fingerprint: Crypto.PublicKey.RSA._RSAobj} dictionary
_server_keys = {}


def get_byte_array(integer):
    """Return the variable length bytes corresponding to the given int"""
    # Operate in big endian (unlike most of Telegram API) since:
    # > "...pq is a representation of a natural number
    #    (in binary *big endian* format)..."
    # > "...current value of dh_prime equals
    #    (in *big-endian* byte order)..."
    # Reference: https://core.telegram.org/mtproto/auth_key
    return int.to_bytes(
        integer,
        length=(integer.bit_length() + 8 - 1) // 8,  # 8 bits per byte,
        byteorder='big',
        signed=False
    )


def _compute_fingerprint(key):
    """For a given Crypto.RSA key, computes its 8-bytes-long fingerprint
       in the same way that Telegram does.
    """
    with BinaryWriter() as writer:
        writer.tgwrite_bytes(get_byte_array(key.n))
        writer.tgwrite_bytes(get_byte_array(key.e))
        # Telegram uses the last 8 bytes as the fingerprint
        return sha1(writer.get_bytes()).digest()[-8:]


def add_key(pub):
    """Adds a new public key to be used when encrypting new data is needed"""
    global _server_keys
    key = rsa.PublicKey.load_pkcs1(pub)
    _server_keys[_compute_fingerprint(key)] = key


def encrypt(fingerprint, data):
    """Given the fingerprint of a previously added RSA key, encrypt its data
       in the way Telegram requires us to do so (sha1(data) + data + padding)
    """
    global _server_keys
    key = _server_keys.get(fingerprint, None)
    if not key:
        return None

    # len(sha1.digest) is always 20, so we're left with 255 - 20 - x padding
    to_encrypt = sha1(data).digest() + data + os.urandom(235 - len(data))

    # rsa module rsa.encrypt adds 11 bits for padding which we don't want
    # rsa module uses rsa.transform.bytes2int(to_encrypt), easier way:
    payload = int.from_bytes(to_encrypt, 'big')
    encrypted = rsa.core.encrypt_int(payload, key.e, key.n)
    # rsa module uses transform.int2bytes(encrypted, keylength), easier:
    block = encrypted.to_bytes(256, 'big')
    return block


# Add default keys
for pub in (
        '''-----BEGIN RSA PUBLIC KEY-----
MIIBCgKCAQEAwVACPi9w23mF3tBkdZz+zwrzKOaaQdr01vAbU4E1pvkfj4sqDsm6
lyDONS789sVoD/xCS9Y0hkkC3gtL1tSfTlgCMOOul9lcixlEKzwKENj1Yz/s7daS
an9tqw3bfUV/nqgbhGX81v/+7RFAEd+RwFnK7a+XYl9sluzHRyVVaTTveB2GazTw
Efzk2DWgkBluml8OREmvfraX3bkHZJTKX4EQSjBbbdJ2ZXIsRrYOXfaA+xayEGB+
8hdlLmAjbCVfaigxX0CDqWeR1yFL9kwd9P0NsZRPsmoqVwMbMu7mStFai6aIhc3n
Slv8kg9qv1m6XHVQY3PnEw+QQtqSIXklHwIDAQAB
-----END RSA PUBLIC KEY-----''',

        '''-----BEGIN RSA PUBLIC KEY-----
MIIBCgKCAQEAxq7aeLAqJR20tkQQMfRn+ocfrtMlJsQ2Uksfs7Xcoo77jAid0bRt
ksiVmT2HEIJUlRxfABoPBV8wY9zRTUMaMA654pUX41mhyVN+XoerGxFvrs9dF1Ru
vCHbI02dM2ppPvyytvvMoefRoL5BTcpAihFgm5xCaakgsJ/tH5oVl74CdhQw8J5L
xI/K++KJBUyZ26Uba1632cOiq05JBUW0Z2vWIOk4BLysk7+U9z+SxynKiZR3/xdi
XvFKk01R3BHV+GUKM2RYazpS/P8v7eyKhAbKxOdRcFpHLlVwfjyM1VlDQrEZxsMp
NTLYXb6Sce1Uov0YtNx5wEowlREH1WOTlwIDAQAB
-----END RSA PUBLIC KEY-----''',

        '''-----BEGIN RSA PUBLIC KEY-----
MIIBCgKCAQEAsQZnSWVZNfClk29RcDTJQ76n8zZaiTGuUsi8sUhW8AS4PSbPKDm+
DyJgdHDWdIF3HBzl7DHeFrILuqTs0vfS7Pa2NW8nUBwiaYQmPtwEa4n7bTmBVGsB
1700/tz8wQWOLUlL2nMv+BPlDhxq4kmJCyJfgrIrHlX8sGPcPA4Y6Rwo0MSqYn3s
g1Pu5gOKlaT9HKmE6wn5Sut6IiBjWozrRQ6n5h2RXNtO7O2qCDqjgB2vBxhV7B+z
hRbLbCmW0tYMDsvPpX5M8fsO05svN+lKtCAuz1leFns8piZpptpSCFn7bWxiA9/f
x5x17D7pfah3Sy2pA+NDXyzSlGcKdaUmwQIDAQAB
-----END RSA PUBLIC KEY-----''',

        '''-----BEGIN RSA PUBLIC KEY-----
MIIBCgKCAQEAwqjFW0pi4reKGbkc9pK83Eunwj/k0G8ZTioMMPbZmW99GivMibwa
xDM9RDWabEMyUtGoQC2ZcDeLWRK3W8jMP6dnEKAlvLkDLfC4fXYHzFO5KHEqF06i
qAqBdmI1iBGdQv/OQCBcbXIWCGDY2AsiqLhlGQfPOI7/vvKc188rTriocgUtoTUc
/n/sIUzkgwTqRyvWYynWARWzQg0I9olLBBC2q5RQJJlnYXZwyTL3y9tdb7zOHkks
WV9IMQmZmyZh/N7sMbGWQpt4NMchGpPGeJ2e5gHBjDnlIf2p1yZOYeUYrdbwcS0t
UiggS4UeE8TzIuXFQxw7fzEIlmhIaq3FnwIDAQAB
-----END RSA PUBLIC KEY-----'''
):
    add_key(pub)



"""File generated by TLObjects' generator. All changes will be ERASED"""
from ...tl.tlobject import TLObject


class GetCdnFileRequest(TLObject):
    constructor_id = 0x2000bcc3
    subclass_of_id = 0xf5ccf928

    def __init__(self, file_token, offset, limit):
        """
        :param file_token: Telegram type: "bytes".
        :param offset: Telegram type: "int".
        :param limit: Telegram type: "int".

        :returns upload.CdnFile: Instance of either CdnFileReuploadNeeded, CdnFile.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.file_token = file_token
        self.offset = offset
        self.limit = limit

    def to_dict(self):
        return {
            'file_token': self.file_token,
            'offset': self.offset,
            'limit': self.limit,
        }

    def on_send(self, writer):
        writer.write_int(GetCdnFileRequest.constructor_id, signed=False)
        writer.tgwrite_bytes(self.file_token)
        writer.write_int(self.offset)
        writer.write_int(self.limit)

    @staticmethod
    def empty():
        return GetCdnFileRequest(None, None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'upload.getCdnFile#2000bcc3 file_token:bytes offset:int limit:int = upload.CdnFile'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class GetCdnFileHashesRequest(TLObject):
    constructor_id = 0xf715c87b
    subclass_of_id = 0x7ec58e40

    def __init__(self, file_token, offset):
        """
        :param file_token: Telegram type: "bytes".
        :param offset: Telegram type: "int".

        :returns Vector<CdnFileHash>: This type has no constructors.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.file_token = file_token
        self.offset = offset

    def to_dict(self):
        return {
            'file_token': self.file_token,
            'offset': self.offset,
        }

    def on_send(self, writer):
        writer.write_int(GetCdnFileHashesRequest.constructor_id, signed=False)
        writer.tgwrite_bytes(self.file_token)
        writer.write_int(self.offset)

    @staticmethod
    def empty():
        return GetCdnFileHashesRequest(None, None)

    def on_response(self, reader):
        self.result = reader.tgread_vector()

    def __repr__(self):
        return 'upload.getCdnFileHashes#f715c87b file_token:bytes offset:int = Vector<CdnFileHash>'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class GetFileRequest(TLObject):
    constructor_id = 0xe3a6cfb5
    subclass_of_id = 0x6c9bd728

    def __init__(self, location, offset, limit):
        """
        :param location: Telegram type: "InputFileLocation".
        :param offset: Telegram type: "int".
        :param limit: Telegram type: "int".

        :returns upload.File: Instance of either File, FileCdnRedirect.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.location = location
        self.offset = offset
        self.limit = limit

    def to_dict(self):
        return {
            'location': None if self.location is None else self.location.to_dict(),
            'offset': self.offset,
            'limit': self.limit,
        }

    def on_send(self, writer):
        writer.write_int(GetFileRequest.constructor_id, signed=False)
        self.location.on_send(writer)
        writer.write_int(self.offset)
        writer.write_int(self.limit)

    @staticmethod
    def empty():
        return GetFileRequest(None, None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'upload.getFile#e3a6cfb5 location:InputFileLocation offset:int limit:int = upload.File'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class GetWebFileRequest(TLObject):
    constructor_id = 0x24e6818d
    subclass_of_id = 0x68f17f51

    def __init__(self, location, offset, limit):
        """
        :param location: Telegram type: "InputWebFileLocation".
        :param offset: Telegram type: "int".
        :param limit: Telegram type: "int".

        :returns upload.WebFile: Instance of WebFile.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.location = location
        self.offset = offset
        self.limit = limit

    def to_dict(self):
        return {
            'location': None if self.location is None else self.location.to_dict(),
            'offset': self.offset,
            'limit': self.limit,
        }

    def on_send(self, writer):
        writer.write_int(GetWebFileRequest.constructor_id, signed=False)
        self.location.on_send(writer)
        writer.write_int(self.offset)
        writer.write_int(self.limit)

    @staticmethod
    def empty():
        return GetWebFileRequest(None, None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'upload.getWebFile#24e6818d location:InputWebFileLocation offset:int limit:int = upload.WebFile'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class ReuploadCdnFileRequest(TLObject):
    constructor_id = 0x1af91c09
    subclass_of_id = 0x7ec58e40

    def __init__(self, file_token, request_token):
        """
        :param file_token: Telegram type: "bytes".
        :param request_token: Telegram type: "bytes".

        :returns Vector<CdnFileHash>: This type has no constructors.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.file_token = file_token
        self.request_token = request_token

    def to_dict(self):
        return {
            'file_token': self.file_token,
            'request_token': self.request_token,
        }

    def on_send(self, writer):
        writer.write_int(ReuploadCdnFileRequest.constructor_id, signed=False)
        writer.tgwrite_bytes(self.file_token)
        writer.tgwrite_bytes(self.request_token)

    @staticmethod
    def empty():
        return ReuploadCdnFileRequest(None, None)

    def on_response(self, reader):
        self.result = reader.tgread_vector()

    def __repr__(self):
        return 'upload.reuploadCdnFile#1af91c09 file_token:bytes request_token:bytes = Vector<CdnFileHash>'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class SaveBigFilePartRequest(TLObject):
    constructor_id = 0xde7b673d
    subclass_of_id = 0xf5b399ac

    def __init__(self, file_id, file_part, file_total_parts, bytes):
        """
        :param file_id: Telegram type: "long".
        :param file_part: Telegram type: "int".
        :param file_total_parts: Telegram type: "int".
        :param bytes: Telegram type: "bytes".

        :returns Bool: This type has no constructors.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.file_id = file_id
        self.file_part = file_part
        self.file_total_parts = file_total_parts
        self.bytes = bytes

    def to_dict(self):
        return {
            'file_id': self.file_id,
            'file_part': self.file_part,
            'file_total_parts': self.file_total_parts,
            'bytes': self.bytes,
        }

    def on_send(self, writer):
        writer.write_int(SaveBigFilePartRequest.constructor_id, signed=False)
        writer.write_long(self.file_id)
        writer.write_int(self.file_part)
        writer.write_int(self.file_total_parts)
        writer.tgwrite_bytes(self.bytes)

    @staticmethod
    def empty():
        return SaveBigFilePartRequest(None, None, None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'upload.saveBigFilePart#de7b673d file_id:long file_part:int file_total_parts:int bytes:bytes = Bool'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class SaveFilePartRequest(TLObject):
    constructor_id = 0xb304a621
    subclass_of_id = 0xf5b399ac

    def __init__(self, file_id, file_part, bytes):
        """
        :param file_id: Telegram type: "long".
        :param file_part: Telegram type: "int".
        :param bytes: Telegram type: "bytes".

        :returns Bool: This type has no constructors.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.file_id = file_id
        self.file_part = file_part
        self.bytes = bytes

    def to_dict(self):
        return {
            'file_id': self.file_id,
            'file_part': self.file_part,
            'bytes': self.bytes,
        }

    def on_send(self, writer):
        writer.write_int(SaveFilePartRequest.constructor_id, signed=False)
        writer.write_long(self.file_id)
        writer.write_int(self.file_part)
        writer.tgwrite_bytes(self.bytes)

    @staticmethod
    def empty():
        return SaveFilePartRequest(None, None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'upload.saveFilePart#b304a621 file_id:long file_part:int bytes:bytes = Bool'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)




import os
from datetime import datetime
from io import BufferedReader, BytesIO
from struct import unpack

from ..errors import InvalidParameterError, TypeNotFoundError
from ..tl.all_tlobjects import tlobjects


class BinaryReader:
    """
    Small utility class to read binary data.
    Also creates a "Memory Stream" if necessary
    """

    def __init__(self, data=None, stream=None):
        if data:
            self.stream = BytesIO(data)
        elif stream:
            self.stream = stream
        else:
            raise InvalidParameterError(
                'Either bytes or a stream must be provided')

        self.reader = BufferedReader(self.stream)

    # region Reading

    # "All numbers are written as little endian."
    # https://core.telegram.org/mtproto
    def read_byte(self):
        """Reads a single byte value"""
        return self.read(1)[0]

    def read_int(self, signed=True):
        """Reads an integer (4 bytes) value"""
        return int.from_bytes(self.read(4), byteorder='little', signed=signed)

    def read_long(self, signed=True):
        """Reads a long integer (8 bytes) value"""
        return int.from_bytes(self.read(8), byteorder='little', signed=signed)

    def read_float(self):
        """Reads a real floating point (4 bytes) value"""
        return unpack('<f', self.read(4))[0]

    def read_double(self):
        """Reads a real floating point (8 bytes) value"""
        return unpack('<d', self.read(8))[0]

    def read_large_int(self, bits, signed=True):
        """Reads a n-bits long integer value"""
        return int.from_bytes(
            self.read(bits // 8), byteorder='little', signed=signed)

    def read(self, length):
        """Read the given amount of bytes"""
        result = self.reader.read(length)
        if len(result) != length:
            raise BufferError('No more data left to read')

        return result

    def get_bytes(self):
        """Gets the byte array representing the current buffer as a whole"""
        return self.stream.getvalue()

    # endregion

    # region Telegram custom reading

    def tgread_bytes(self):
        """Reads a Telegram-encoded byte array,
           without the need of specifying its length
        """
        first_byte = self.read_byte()
        if first_byte == 254:
            length = self.read_byte() | (self.read_byte() << 8) | (
                self.read_byte() << 16)
            padding = length % 4
        else:
            length = first_byte
            padding = (length + 1) % 4

        data = self.read(length)
        if padding > 0:
            padding = 4 - padding
            self.read(padding)

        return data

    def tgread_string(self):
        """Reads a Telegram-encoded string"""
        return str(self.tgread_bytes(), encoding='utf-8', errors='replace')

    def tgread_bool(self):
        """Reads a Telegram boolean value"""
        value = self.read_int(signed=False)
        if value == 0x997275b5:  # boolTrue
            return True
        elif value == 0xbc799737:  # boolFalse
            return False
        else:
            raise ValueError('Invalid boolean code {}'.format(hex(value)))

    def tgread_date(self):
        """Reads and converts Unix time (used by Telegram)
           into a Python datetime object
        """
        value = self.read_int()
        return None if value == 0 else datetime.fromtimestamp(value)

    def tgread_object(self):
        """Reads a Telegram object"""
        constructor_id = self.read_int(signed=False)
        clazz = tlobjects.get(constructor_id, None)
        if clazz is None:
            # The class was None, but there's still a
            # chance of it being a manually parsed value like bool!
            value = constructor_id
            if value == 0x997275b5:  # boolTrue
                return True
            elif value == 0xbc799737:  # boolFalse
                return False

            # If there was still no luck, give up
            raise TypeNotFoundError(constructor_id)

        # Create an empty instance of the class and
        # fill it with the read attributes
        result = clazz.empty()
        result.on_response(self)
        return result

    def tgread_vector(self):
        """Reads a vector (a list) of Telegram objects"""
        if 0x1cb5c415 != self.read_int(signed=False):
            raise ValueError('Invalid constructor code, vector was expected')

        count = self.read_int()
        return [self.tgread_object() for _ in range(count)]

    # endregion

    def close(self):
        self.reader.close()

    # region Position related

    def tell_position(self):
        """Tells the current position on the stream"""
        return self.reader.tell()

    def set_position(self, position):
        """Sets the current position on the stream"""
        self.reader.seek(position)

    def seek(self, offset):
        """Seeks the stream position given an offset from the
           current position. The offset may be negative
        """
        self.reader.seek(offset, os.SEEK_CUR)

    # endregion

    # region with block

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    # endregion



import logging
from datetime import timedelta
from hashlib import md5
from io import BytesIO
from os import path

from . import helpers as utils
from .crypto import rsa, CdnDecrypter
from .errors import (
    RPCError, FloodWaitError, FileMigrateError, TypeNotFoundError
)
from .network import authenticator, MtProtoSender, Connection, ConnectionMode
from .tl import TLObject, Session
from .tl.all_tlobjects import layer
from .tl.functions import (
    InitConnectionRequest, InvokeWithLayerRequest, PingRequest
)
from .tl.functions.auth import (
    ImportAuthorizationRequest, ExportAuthorizationRequest
)
from .tl.functions.help import (
    GetCdnConfigRequest, GetConfigRequest
)
from .tl.functions.upload import (
    GetFileRequest, SaveBigFilePartRequest, SaveFilePartRequest
)
from .tl.types import InputFile, InputFileBig
from .tl.types.upload import FileCdnRedirect
from .update_state import UpdateState
from .utils import get_appropriated_part_size


class TelegramBareClient:
    """Bare Telegram Client with just the minimum -

       The reason to distinguish between a MtProtoSender and a
       TelegramClient itself is because the sender is just that,
       a sender, which should know nothing about Telegram but
       rather how to handle this specific connection.

       The TelegramClient itself should know how to initialize
       a proper connection to the servers, as well as other basic
       methods such as disconnection and reconnection.

       This distinction between a bare client and a full client
       makes it possible to create clones of the bare version
       (by using the same session, IP address and port) to be
       able to execute queries on either, without the additional
       cost that would involve having the methods for signing in,
       logging out, and such.
    """

    # Current TelegramClient version
    __version__ = '0.13.2.1'

    # region Initialization

    def __init__(self, session, api_id, api_hash,
                 connection_mode=ConnectionMode.TCP_FULL,
                 proxy=None,
                 process_updates=False,
                 timeout=timedelta(seconds=5)):
        """Initializes the Telegram client with the specified API ID and Hash.
           Session must always be a Session instance, and an optional proxy
           can also be specified to be used on the connection.
        """
        self.session = session
        self.api_id = int(api_id)
        self.api_hash = api_hash
        self._connection_mode = connection_mode
        self.proxy = proxy
        self._timeout = timeout
        self._logger = logging.getLogger(__name__)

        # Cache "exported" senders 'dc_id: TelegramBareClient' and
        # their corresponding sessions not to recreate them all
        # the time since it's a (somewhat expensive) process.
        self._cached_clients = {}

        # This member will process updates if enabled.
        # One may change self.updates.enabled at any later point.
        self.updates = UpdateState(process_updates)

        # These will be set later
        self.dc_options = None
        self._sender = None

    # endregion

    # region Connecting

    def connect(self, exported_auth=None, initial_query=None):
        """Connects to the Telegram servers, executing authentication if
           required. Note that authenticating to the Telegram servers is
           not the same as authenticating the desired user itself, which
           may require a call (or several) to 'sign_in' for the first time.

           If 'exported_auth' is not None, it will be used instead to
           determine the authorization key for the current session.

           If 'initial_query' is not None, it will override the default
           'GetConfigRequest()', and its result will be returned ONLY
           if the client wasn't connected already.
        """
        if self._sender and self._sender.is_connected():
            # Try sending a ping to make sure we're connected already
            # TODO Maybe there's a better way to check this
            try:
                if initial_query is None:
                    self(PingRequest(utils.generate_random_long()))
                    return True
                else:
                    return self(initial_query)
            except:
                # If ping failed, ensure we're disconnected first
                self.disconnect()

        connection = Connection(
            self.session.server_address, self.session.port,
            mode=self._connection_mode, proxy=self.proxy, timeout=self._timeout
        )

        try:
            if not self.session.auth_key:
                self.session.auth_key, self.session.time_offset = \
                    authenticator.do_authentication(connection)

                self.session.save()

            self._sender = MtProtoSender(connection, self.session)
            self._sender.connect()

            # Now it's time to send an InitConnectionRequest
            # This must always be invoked with the layer we'll be using
            if exported_auth is None:
                query = initial_query if initial_query else GetConfigRequest()
            else:
                query = ImportAuthorizationRequest(
                    exported_auth.id, exported_auth.bytes)

            request = InitConnectionRequest(
                api_id=self.api_id,
                device_model=self.session.device_model,
                system_version=self.session.system_version,
                app_version=self.session.app_version,
                lang_code=self.session.lang_code,
                system_lang_code=self.session.system_lang_code,
                lang_pack='',  # "langPacks are for official apps only"
                query=query)

            result = self(InvokeWithLayerRequest(
                layer=layer, query=request
            ))

            if initial_query is None:
                if exported_auth is not None:
                    result = self(GetConfigRequest())

                # We're only interested in the DC options,
                # although many other options are available!
                self.dc_options = result.dc_options
                return True
            else:
                return result

        except TypeNotFoundError as e:
            # This is fine, probably layer migration
            self._logger.debug('Found invalid item, probably migrating', e)
            self.disconnect()
            return self.connect(exported_auth=exported_auth,
                                initial_query=initial_query)

        except (RPCError, ConnectionError) as error:
            # Probably errors from the previous session, ignore them
            self.disconnect()
            self._logger.debug(
                'Could not stabilise initial connection: {}'.format(error)
            )
            return None if initial_query else False

    def disconnect(self):
        """Disconnects from the Telegram server"""
        if self._sender:
            self._sender.disconnect()
            self._sender = None

    def reconnect(self, new_dc=None):
        """Disconnects and connects again (effectively reconnecting).

           If 'new_dc' is not None, the current authorization key is
           removed, the DC used is switched, and a new connection is made.
        """
        self.disconnect()

        if new_dc is not None:
            self.session.auth_key = None  # Force creating new auth_key
            dc = self._get_dc(new_dc)
            self.session.server_address = dc.ip_address
            self.session.port = dc.port
            self.session.save()

        self.connect()

    # endregion

    # region Working with different Data Centers

    def _get_dc(self, dc_id, ipv6=False, cdn=False):
        """Gets the Data Center (DC) associated to 'dc_id'"""
        if not self.dc_options:
            raise ConnectionError(
                'Cannot determine the required data center IP address. '
                'Stabilise a successful initial connection first.')

        try:
            if cdn:
                # Ensure we have the latest keys for the CDNs
                for pk in self(GetCdnConfigRequest()).public_keys:
                    rsa.add_key(pk.public_key)

            return next(
                dc for dc in self.dc_options if dc.id == dc_id and
                bool(dc.ipv6) == ipv6 and bool(dc.cdn) == cdn
            )
        except StopIteration:
            if not cdn:
                raise

            # New configuration, perhaps a new CDN was added?
            self.dc_options = self(GetConfigRequest()).dc_options
            return self._get_dc(dc_id, ipv6=ipv6, cdn=cdn)

    def _get_exported_client(self, dc_id,
                             init_connection=False,
                             bypass_cache=False):
        """Gets a cached exported TelegramBareClient for the desired DC.

           If it's the first time retrieving the TelegramBareClient, the
           current authorization is exported to the new DC so that
           it can be used there, and the connection is initialized.

           If after using the sender a ConnectionResetError is raised,
           this method should be called again with init_connection=True
           in order to perform the reconnection.

           If bypass_cache is True, a new client will be exported and
           it will not be cached.
        """
        # Thanks badoualy/kotlogram on /telegram/api/DefaultTelegramClient.kt
        # for clearly showing how to export the authorization! ^^
        client = self._cached_clients.get(dc_id)
        if client and not bypass_cache:
            if init_connection:
                client.reconnect()
            return client
        else:
            dc = self._get_dc(dc_id)

            # Export the current authorization to the new DC.
            export_auth = self(ExportAuthorizationRequest(dc_id))

            # Create a temporary session for this IP address, which needs
            # to be different because each auth_key is unique per DC.
            #
            # Construct this session with the connection parameters
            # (system version, device model...) from the current one.
            session = Session(self.session)
            session.server_address = dc.ip_address
            session.port = dc.port
            client = TelegramBareClient(
                session, self.api_id, self.api_hash,
                timeout=self._timeout
            )
            client.connect(exported_auth=export_auth)

            if not bypass_cache:
                # Don't go through this expensive process every time.
                self._cached_clients[dc_id] = client
            return client

    # endregion

    # region Invoking Telegram requests

    def invoke(self, request, updates=None, call_receive=True):
        """Invokes (sends) a MTProtoRequest and returns (receives) its result.

           If 'updates' is not None, all read update object will be put
           in such list. Otherwise, update objects will be ignored.

           If 'call_receive' is set to False, then there should be another
           thread calling to 'self._sender.receive()' running or this method
           will lock forever.
        """
        if not isinstance(request, TLObject) and not request.content_related:
            raise ValueError('You can only invoke requests, not types!')

        if not self._sender:
            raise ValueError('You must be connected to invoke requests!')

        try:
            # Ensure that we start with no previous errors (i.e. resending)
            request.confirm_received.clear()
            request.rpc_error = None

            self._sender.send(request)
            if not call_receive:
                # TODO This will be slightly troublesome if we allow
                # switching between constant read or not on the fly.
                # Must also watch out for calling .read() from two places,
                # in which case a Lock would be required for .receive().
                request.confirm_received.wait()  # TODO Socket's timeout here?
            else:
                while not request.confirm_received.is_set():
                    self._sender.receive(update_state=self.updates)

            if request.rpc_error:
                raise request.rpc_error
            return request.result

        except ConnectionResetError:
            self._logger.debug('Server disconnected us. Reconnecting and '
                               'resending request...')
            self.reconnect()
            return self.invoke(request)

        except FloodWaitError:
            self.disconnect()
            raise

    # Let people use client(SomeRequest()) instead client.invoke(...)
    __call__ = invoke

    # endregion

    # region Uploading media

    def upload_file(self,
                    file,
                    part_size_kb=None,
                    file_name=None,
                    progress_callback=None):
        """Uploads the specified file and returns a handle (an instance
           of InputFile or InputFileBig, as required) which can be later used.

           Uploading a file will simply return a "handle" to the file stored
           remotely in the Telegram servers, which can be later used on. This
           will NOT upload the file to your own chat.

           'file' may be either a file path, a byte array, or a stream.
           Note that if the file is a stream it will need to be read
           entirely into memory to tell its size first.

           If 'progress_callback' is not None, it should be a function that
           takes two parameters, (bytes_uploaded, total_bytes).

           Default values for the optional parameters if left as None are:
             part_size_kb = get_appropriated_part_size(file_size)
             file_name    = path.basename(file_path)
        """
        if isinstance(file, str):
            file_size = path.getsize(file)
        elif isinstance(file, bytes):
            file_size = len(file)
        else:
            file = file.read()
            file_size = len(file)

        if not part_size_kb:
            part_size_kb = get_appropriated_part_size(file_size)

        if part_size_kb > 512:
            raise ValueError('The part size must be less or equal to 512KB')

        part_size = int(part_size_kb * 1024)
        if part_size % 1024 != 0:
            raise ValueError('The part size must be evenly divisible by 1024')

        # Determine whether the file is too big (over 10MB) or not
        # Telegram does make a distinction between smaller or larger files
        is_large = file_size > 10 * 1024 * 1024
        part_count = (file_size + part_size - 1) // part_size

        file_id = utils.generate_random_long()
        hash_md5 = md5()

        stream = open(file, 'rb') if isinstance(file, str) else BytesIO(file)
        try:
            for part_index in range(part_count):
                # Read the file by in chunks of size part_size
                part = stream.read(part_size)

                # The SavePartRequest is different depending on whether
                # the file is too large or not (over or less than 10MB)
                if is_large:
                    request = SaveBigFilePartRequest(file_id, part_index,
                                                     part_count, part)
                else:
                    request = SaveFilePartRequest(file_id, part_index, part)

                result = self(request)
                if result:
                    if not is_large:
                        # No need to update the hash if it's a large file
                        hash_md5.update(part)

                    if progress_callback:
                        progress_callback(stream.tell(), file_size)
                else:
                    raise ValueError('Failed to upload file part {}.'
                                     .format(part_index))
        finally:
            stream.close()

        # Set a default file name if None was specified
        if not file_name:
            if isinstance(file, str):
                file_name = path.basename(file)
            else:
                file_name = str(file_id)

        if is_large:
            return InputFileBig(file_id, part_count, file_name)
        else:
            return InputFile(file_id, part_count, file_name,
                             md5_checksum=hash_md5.hexdigest())

    # endregion

    # region Downloading media

    def download_file(self,
                      input_location,
                      file,
                      part_size_kb=None,
                      file_size=None,
                      progress_callback=None):
        """Downloads the given InputFileLocation to file (a stream or str).

           If 'progress_callback' is not None, it should be a function that
           takes two parameters, (bytes_downloaded, total_bytes). Note that
           'total_bytes' simply equals 'file_size', and may be None.
        """
        if not part_size_kb:
            if not file_size:
                part_size_kb = 64  # Reasonable default
            else:
                part_size_kb = get_appropriated_part_size(file_size)

        part_size = int(part_size_kb * 1024)
        # https://core.telegram.org/api/files says:
        # > part_size % 1024 = 0 (divisible by 1KB)
        #
        # But https://core.telegram.org/cdn (more recent) says:
        # > limit must be divisible by 4096 bytes
        # So we just stick to the 4096 limit.
        if part_size % 4096 != 0:
            raise ValueError('The part size must be evenly divisible by 4096.')

        if isinstance(file, str):
            # Ensure that we'll be able to download the media
            utils.ensure_parent_dir_exists(file)
            f = open(file, 'wb')
        else:
            f = file

        # The used client will change if FileMigrateError occurs
        client = self
        cdn_decrypter = None

        try:
            offset_index = 0
            while True:
                offset = offset_index * part_size

                try:
                    if cdn_decrypter:
                        result = cdn_decrypter.get_file()
                    else:
                        result = client(GetFileRequest(
                            input_location, offset, part_size
                        ))

                        if isinstance(result, FileCdnRedirect):
                            cdn_decrypter, result = \
                                CdnDecrypter.prepare_decrypter(
                                    client, TelegramBareClient, result
                                )

                except FileMigrateError as e:
                    client = self._get_exported_client(e.new_dc)
                    continue

                offset_index += 1

                # If we have received no data (0 bytes), the file is over
                # So there is nothing left to download and write
                if not result.bytes:
                    if cdn_decrypter:
                        cdn_decrypter.finish_check()

                    # Return some extra information, unless it's a CDN file
                    return getattr(result, 'type', '')

                f.write(result.bytes)
                if progress_callback:
                    progress_callback(f.tell(), file_size)
        finally:
            if cdn_decrypter:
                try:
                    cdn_decrypter.client.disconnect()
                except:
                    pass
            if isinstance(file, str):
                f.close()

    # endregion



"""File generated by TLObjects' generator. All changes will be ERASED"""
from ...tl.tlobject import TLObject
from ...utils import get_input_user


class GetFullUserRequest(TLObject):
    constructor_id = 0xca30a5b1
    subclass_of_id = 0x1f4661b9

    def __init__(self, id):
        """
        :param id: Telegram type: "InputUser".

        :returns UserFull: Instance of UserFull.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.id = get_input_user(id)

    def to_dict(self):
        return {
            'id': None if self.id is None else self.id.to_dict(),
        }

    def on_send(self, writer):
        writer.write_int(GetFullUserRequest.constructor_id, signed=False)
        self.id.on_send(writer)

    @staticmethod
    def empty():
        return GetFullUserRequest(None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'users.getFullUser#ca30a5b1 id:InputUser = UserFull'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)

from ...utils import get_input_user


class GetUsersRequest(TLObject):
    constructor_id = 0xd91a548
    subclass_of_id = 0x406da4d

    def __init__(self, id):
        """
        :param id: Telegram type: "InputUser". Must be a list.

        :returns Vector<User>: This type has no constructors.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.id = [get_input_user(_x) for _x in id]

    def to_dict(self):
        return {
            'id': [] if self.id is None else [None if x is None else x.to_dict() for x in self.id],
        }

    def on_send(self, writer):
        writer.write_int(GetUsersRequest.constructor_id, signed=False)
        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.id))
        for _x in self.id:
            _x.on_send(writer)

    @staticmethod
    def empty():
        return GetUsersRequest(None)

    def on_response(self, reader):
        self.result = reader.tgread_vector()

    def __repr__(self):
        return 'users.getUsers#0d91a548 id:Vector<InputUser> = Vector<User>'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)




"""File generated by TLObjects' generator. All changes will be ERASED"""
from ...tl.tlobject import TLObject


class AppUpdate(TLObject):
    constructor_id = 0x8987f311
    subclass_of_id = 0x5897069e

    def __init__(self, id, critical, url, text):
        """
        :param id: Telegram type: "int".
        :param critical: Telegram type: "Bool".
        :param url: Telegram type: "string".
        :param text: Telegram type: "string".

        Constructor for help.AppUpdate: Instance of either AppUpdate, NoAppUpdate.
        """
        super().__init__()

        self.id = id
        self.critical = critical
        self.url = url
        self.text = text

    def to_dict(self):
        return {
            'id': self.id,
            'critical': self.critical,
            'url': self.url,
            'text': self.text,
        }

    def on_send(self, writer):
        writer.write_int(AppUpdate.constructor_id, signed=False)
        writer.write_int(self.id)
        writer.tgwrite_bool(self.critical)
        writer.tgwrite_string(self.url)
        writer.tgwrite_string(self.text)

    @staticmethod
    def empty():
        return AppUpdate(None, None, None, None)

    def on_response(self, reader):
        self.id = reader.read_int()
        self.critical = reader.tgread_bool()
        self.url = reader.tgread_string()
        self.text = reader.tgread_string()

    def __repr__(self):
        return 'help.appUpdate#8987f311 id:int critical:Bool url:string text:string = help.AppUpdate'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class ConfigSimple(TLObject):
    constructor_id = 0xd997c3c5
    subclass_of_id = 0x29183ac4

    def __init__(self, date, expires, dc_id, ip_port_list):
        """
        :param date: Telegram type: "date".
        :param expires: Telegram type: "date".
        :param dc_id: Telegram type: "int".
        :param ip_port_list: Telegram type: "ipPort". Must be a list.

        Constructor for help.ConfigSimple: Instance of ConfigSimple.
        """
        super().__init__()

        self.date = date
        self.expires = expires
        self.dc_id = dc_id
        self.ip_port_list = ip_port_list

    def to_dict(self):
        return {
            'date': self.date,
            'expires': self.expires,
            'dc_id': self.dc_id,
            'ip_port_list': [] if self.ip_port_list is None else [None if x is None else x.to_dict() for x in self.ip_port_list],
        }

    def on_send(self, writer):
        writer.write_int(ConfigSimple.constructor_id, signed=False)
        writer.tgwrite_date(self.date)
        writer.tgwrite_date(self.expires)
        writer.write_int(self.dc_id)
        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.ip_port_list))
        for _x in self.ip_port_list:
            _x.on_send(writer)

    @staticmethod
    def empty():
        return ConfigSimple(None, None, None, None)

    def on_response(self, reader):
        self.date = reader.tgread_date()
        self.expires = reader.tgread_date()
        self.dc_id = reader.read_int()
        reader.read_int()
        self.ip_port_list = []
        _len = reader.read_int()
        for _ in range(_len):
            _x = reader.tgread_object()
            self.ip_port_list.append(_x)

    def __repr__(self):
        return 'help.configSimple#d997c3c5 date:int expires:int dc_id:int ip_port_list:Vector<ipPort> = help.ConfigSimple'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class InviteText(TLObject):
    constructor_id = 0x18cb9f78
    subclass_of_id = 0xcf70aa35

    def __init__(self, message):
        """
        :param message: Telegram type: "string".

        Constructor for help.InviteText: Instance of InviteText.
        """
        super().__init__()

        self.message = message

    def to_dict(self):
        return {
            'message': self.message,
        }

    def on_send(self, writer):
        writer.write_int(InviteText.constructor_id, signed=False)
        writer.tgwrite_string(self.message)

    @staticmethod
    def empty():
        return InviteText(None)

    def on_response(self, reader):
        self.message = reader.tgread_string()

    def __repr__(self):
        return 'help.inviteText#18cb9f78 message:string = help.InviteText'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class NoAppUpdate(TLObject):
    constructor_id = 0xc45a6536
    subclass_of_id = 0x5897069e

    def __init__(self):
        super().__init__()

    @staticmethod
    def to_dict():
        return {}

    def on_send(self, writer):
        writer.write_int(NoAppUpdate.constructor_id, signed=False)

    @staticmethod
    def empty():
        return NoAppUpdate()

    def on_response(self, reader):
        pass

    def __repr__(self):
        return 'help.noAppUpdate#c45a6536 = help.AppUpdate'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class Support(TLObject):
    constructor_id = 0x17c6b5f6
    subclass_of_id = 0x7159bceb

    def __init__(self, phone_number, user):
        """
        :param phone_number: Telegram type: "string".
        :param user: Telegram type: "User".

        Constructor for help.Support: Instance of Support.
        """
        super().__init__()

        self.phone_number = phone_number
        self.user = user

    def to_dict(self):
        return {
            'phone_number': self.phone_number,
            'user': None if self.user is None else self.user.to_dict(),
        }

    def on_send(self, writer):
        writer.write_int(Support.constructor_id, signed=False)
        writer.tgwrite_string(self.phone_number)
        self.user.on_send(writer)

    @staticmethod
    def empty():
        return Support(None, None)

    def on_response(self, reader):
        self.phone_number = reader.tgread_string()
        self.user = reader.tgread_object()

    def __repr__(self):
        return 'help.support#17c6b5f6 phone_number:string user:User = help.Support'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class TermsOfService(TLObject):
    constructor_id = 0xf1ee3e90
    subclass_of_id = 0x20ee8312

    def __init__(self, text):
        """
        :param text: Telegram type: "string".

        Constructor for help.TermsOfService: Instance of TermsOfService.
        """
        super().__init__()

        self.text = text

    def to_dict(self):
        return {
            'text': self.text,
        }

    def on_send(self, writer):
        writer.write_int(TermsOfService.constructor_id, signed=False)
        writer.tgwrite_string(self.text)

    @staticmethod
    def empty():
        return TermsOfService(None)

    def on_response(self, reader):
        self.text = reader.tgread_string()

    def __repr__(self):
        return 'help.termsOfService#f1ee3e90 text:string = help.TermsOfService'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)




from hashlib import sha1

from .. import helpers as utils
from ..extensions import BinaryReader, BinaryWriter


class AuthKey:
    def __init__(self, data):
        self.key = data

        with BinaryReader(sha1(self.key).digest()) as reader:
            self.aux_hash = reader.read_long(signed=False)
            reader.read(4)
            self.key_id = reader.read_long(signed=False)

    def calc_new_nonce_hash(self, new_nonce, number):
        """Calculates the new nonce hash based on
           the current class fields' values
        """
        with BinaryWriter() as writer:
            writer.write(new_nonce)
            writer.write_byte(number)
            writer.write_long(self.aux_hash, signed=False)

            new_nonce_hash = utils.calc_msg_key(writer.get_bytes())
            return new_nonce_hash



"""
Utilities for working with the Telegram API itself (such as handy methods
to convert between an entity like an User, Chat, etc. into its Input version)
"""
from mimetypes import add_type, guess_extension

from .tl import TLObject
from .tl.types import (
    Channel, ChannelForbidden, Chat, ChatEmpty, ChatForbidden, ChatFull,
    ChatPhoto, InputPeerChannel, InputPeerChat, InputPeerUser, InputPeerEmpty,
    MessageMediaDocument, MessageMediaPhoto, PeerChannel, InputChannel,
    UserEmpty, InputUser, InputUserEmpty, InputUserSelf, InputPeerSelf,
    PeerChat, PeerUser, User, UserFull, UserProfilePhoto, Document
)


def get_display_name(entity):
    """Gets the input peer for the given "entity" (user, chat or channel)
       Returns None if it was not found"""
    if isinstance(entity, User):
        if entity.last_name and entity.first_name:
            return '{} {}'.format(entity.first_name, entity.last_name)
        elif entity.first_name:
            return entity.first_name
        elif entity.last_name:
            return entity.last_name
        else:
            return '(No name)'

    if isinstance(entity, Chat) or isinstance(entity, Channel):
        return entity.title

    return '(unknown)'

# For some reason, .webp (stickers' format) is not registered
add_type('image/webp', '.webp')


def get_extension(media):
    """Gets the corresponding extension for any Telegram media"""

    # Photos are always compressed as .jpg by Telegram
    if (isinstance(media, UserProfilePhoto) or isinstance(media, ChatPhoto) or
            isinstance(media, MessageMediaPhoto)):
        return '.jpg'

    # Documents will come with a mime type
    if isinstance(media, MessageMediaDocument):
        if isinstance(media.document, Document):
            if media.document.mime_type == 'application/octet-stream':
                # Octet stream are just bytes, which have no default extension
                return ''
            else:
                extension = guess_extension(media.document.mime_type)
                return extension if extension else ''

    return ''


def _raise_cast_fail(entity, target):
    raise ValueError('Cannot cast {} to any kind of {}.'
                     .format(type(entity).__name__, target))


def get_input_peer(entity):
    """Gets the input peer for the given "entity" (user, chat or channel).
       A ValueError is raised if the given entity isn't a supported type."""
    if entity is None:
        return None

    if not isinstance(entity, TLObject):
        _raise_cast_fail(entity, 'InputPeer')

    if type(entity).subclass_of_id == 0xc91c90b6:  # crc32(b'InputPeer')
        return entity

    if isinstance(entity, User):
        if entity.is_self:
            return InputPeerSelf()
        else:
            return InputPeerUser(entity.id, entity.access_hash)

    if any(isinstance(entity, c) for c in (
            Chat, ChatEmpty, ChatForbidden)):
        return InputPeerChat(entity.id)

    if any(isinstance(entity, c) for c in (
            Channel, ChannelForbidden)):
        return InputPeerChannel(entity.id, entity.access_hash)

    # Less common cases
    if isinstance(entity, UserEmpty):
        return InputPeerEmpty()

    if isinstance(entity, InputUser):
        return InputPeerUser(entity.user_id, entity.access_hash)

    if isinstance(entity, UserFull):
        return get_input_peer(entity.user)

    if isinstance(entity, ChatFull):
        return InputPeerChat(entity.id)

    if isinstance(entity, PeerChat):
        return InputPeerChat(entity.chat_id)

    _raise_cast_fail(entity, 'InputPeer')


def get_input_channel(entity):
    """Similar to get_input_peer, but for InputChannel's alone"""
    if entity is None:
        return None

    if not isinstance(entity, TLObject):
        _raise_cast_fail(entity, 'InputChannel')

    if type(entity).subclass_of_id == 0x40f202fd:  # crc32(b'InputChannel')
        return entity

    if isinstance(entity, Channel) or isinstance(entity, ChannelForbidden):
        return InputChannel(entity.id, entity.access_hash)

    if isinstance(entity, InputPeerChannel):
        return InputChannel(entity.channel_id, entity.access_hash)

    _raise_cast_fail(entity, 'InputChannel')


def get_input_user(entity):
    """Similar to get_input_peer, but for InputUser's alone"""
    if entity is None:
        return None

    if not isinstance(entity, TLObject):
        _raise_cast_fail(entity, 'InputUser')

    if type(entity).subclass_of_id == 0xe669bf46:  # crc32(b'InputUser')
        return entity

    if isinstance(entity, User):
        if entity.is_self:
            return InputUserSelf()
        else:
            return InputUser(entity.id, entity.access_hash)

    if isinstance(entity, UserEmpty):
        return InputUserEmpty()

    if isinstance(entity, UserFull):
        return get_input_user(entity.user)

    if isinstance(entity, InputPeerUser):
        return InputUser(entity.user_id, entity.access_hash)

    _raise_cast_fail(entity, 'InputUser')


def find_user_or_chat(peer, users, chats):
    """Finds the corresponding user or chat given a peer.
       Returns None if it was not found"""
    try:
        if isinstance(peer, PeerUser):
            return next(u for u in users if u.id == peer.user_id)

        elif isinstance(peer, PeerChat):
            return next(c for c in chats if c.id == peer.chat_id)

        elif isinstance(peer, PeerChannel):
            return next(c for c in chats if c.id == peer.channel_id)

    except StopIteration: return

    if isinstance(peer, int):
        try: return next(u for u in users if u.id == peer)
        except StopIteration: pass

        try: return next(c for c in chats if c.id == peer)
        except StopIteration: pass


def get_appropriated_part_size(file_size):
    """Gets the appropriated part size when uploading or downloading files,
       given an initial file size"""
    if file_size <= 1048576:  # 1MB
        return 32
    if file_size <= 10485760:  # 10MB
        return 64
    if file_size <= 393216000:  # 375MB
        return 128
    if file_size <= 786432000:  # 750MB
        return 256
    if file_size <= 1572864000:  # 1500MB
        return 512

    raise ValueError('File size too large')



"""File generated by TLObjects' generator. All changes will be ERASED"""
from ...tl.tlobject import TLObject


class Authorization(TLObject):
    constructor_id = 0xcd050916
    subclass_of_id = 0xb9e04e39

    def __init__(self, user, tmp_sessions=None):
        """
        :param tmp_sessions: Telegram type: "int".
        :param user: Telegram type: "User".

        Constructor for auth.Authorization: Instance of Authorization.
        """
        super().__init__()

        self.tmp_sessions = tmp_sessions
        self.user = user

    def to_dict(self):
        return {
            'tmp_sessions': self.tmp_sessions,
            'user': None if self.user is None else self.user.to_dict(),
        }

    def on_send(self, writer):
        writer.write_int(Authorization.constructor_id, signed=False)
        flags = 0
        flags |= (1 << 0) if self.tmp_sessions else 0
        writer.write_int(flags)

        if self.tmp_sessions:
            writer.write_int(self.tmp_sessions)

        self.user.on_send(writer)

    @staticmethod
    def empty():
        return Authorization(None, None)

    def on_response(self, reader):
        flags = reader.read_int()

        if (flags & (1 << 0)) != 0:
            self.tmp_sessions = reader.read_int()

        self.user = reader.tgread_object()

    def __repr__(self):
        return 'auth.authorization#cd050916 flags:# tmp_sessions:flags.0?int user:User = auth.Authorization'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class CheckedPhone(TLObject):
    constructor_id = 0x811ea28e
    subclass_of_id = 0x99a3d765

    def __init__(self, phone_registered):
        """
        :param phone_registered: Telegram type: "Bool".

        Constructor for auth.CheckedPhone: Instance of CheckedPhone.
        """
        super().__init__()

        self.phone_registered = phone_registered

    def to_dict(self):
        return {
            'phone_registered': self.phone_registered,
        }

    def on_send(self, writer):
        writer.write_int(CheckedPhone.constructor_id, signed=False)
        writer.tgwrite_bool(self.phone_registered)

    @staticmethod
    def empty():
        return CheckedPhone(None)

    def on_response(self, reader):
        self.phone_registered = reader.tgread_bool()

    def __repr__(self):
        return 'auth.checkedPhone#811ea28e phone_registered:Bool = auth.CheckedPhone'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class CodeTypeCall(TLObject):
    constructor_id = 0x741cd3e3
    subclass_of_id = 0xb3f3e401

    def __init__(self):
        super().__init__()

    @staticmethod
    def to_dict():
        return {}

    def on_send(self, writer):
        writer.write_int(CodeTypeCall.constructor_id, signed=False)

    @staticmethod
    def empty():
        return CodeTypeCall()

    def on_response(self, reader):
        pass

    def __repr__(self):
        return 'auth.codeTypeCall#741cd3e3 = auth.CodeType'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class CodeTypeFlashCall(TLObject):
    constructor_id = 0x226ccefb
    subclass_of_id = 0xb3f3e401

    def __init__(self):
        super().__init__()

    @staticmethod
    def to_dict():
        return {}

    def on_send(self, writer):
        writer.write_int(CodeTypeFlashCall.constructor_id, signed=False)

    @staticmethod
    def empty():
        return CodeTypeFlashCall()

    def on_response(self, reader):
        pass

    def __repr__(self):
        return 'auth.codeTypeFlashCall#226ccefb = auth.CodeType'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class CodeTypeSms(TLObject):
    constructor_id = 0x72a3158c
    subclass_of_id = 0xb3f3e401

    def __init__(self):
        super().__init__()

    @staticmethod
    def to_dict():
        return {}

    def on_send(self, writer):
        writer.write_int(CodeTypeSms.constructor_id, signed=False)

    @staticmethod
    def empty():
        return CodeTypeSms()

    def on_response(self, reader):
        pass

    def __repr__(self):
        return 'auth.codeTypeSms#72a3158c = auth.CodeType'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class ExportedAuthorization(TLObject):
    constructor_id = 0xdf969c2d
    subclass_of_id = 0x5fd1ec51

    def __init__(self, id, bytes):
        """
        :param id: Telegram type: "int".
        :param bytes: Telegram type: "bytes".

        Constructor for auth.ExportedAuthorization: Instance of ExportedAuthorization.
        """
        super().__init__()

        self.id = id
        self.bytes = bytes

    def to_dict(self):
        return {
            'id': self.id,
            'bytes': self.bytes,
        }

    def on_send(self, writer):
        writer.write_int(ExportedAuthorization.constructor_id, signed=False)
        writer.write_int(self.id)
        writer.tgwrite_bytes(self.bytes)

    @staticmethod
    def empty():
        return ExportedAuthorization(None, None)

    def on_response(self, reader):
        self.id = reader.read_int()
        self.bytes = reader.tgread_bytes()

    def __repr__(self):
        return 'auth.exportedAuthorization#df969c2d id:int bytes:bytes = auth.ExportedAuthorization'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class PasswordRecovery(TLObject):
    constructor_id = 0x137948a5
    subclass_of_id = 0xfa72d43a

    def __init__(self, email_pattern):
        """
        :param email_pattern: Telegram type: "string".

        Constructor for auth.PasswordRecovery: Instance of PasswordRecovery.
        """
        super().__init__()

        self.email_pattern = email_pattern

    def to_dict(self):
        return {
            'email_pattern': self.email_pattern,
        }

    def on_send(self, writer):
        writer.write_int(PasswordRecovery.constructor_id, signed=False)
        writer.tgwrite_string(self.email_pattern)

    @staticmethod
    def empty():
        return PasswordRecovery(None)

    def on_response(self, reader):
        self.email_pattern = reader.tgread_string()

    def __repr__(self):
        return 'auth.passwordRecovery#137948a5 email_pattern:string = auth.PasswordRecovery'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class SentCode(TLObject):
    constructor_id = 0x5e002502
    subclass_of_id = 0x6ce87081

    def __init__(self, type, phone_code_hash, phone_registered=None, next_type=None, timeout=None):
        """
        :param phone_registered: Telegram type: "true".
        :param type: Telegram type: "auth.SentCodeType".
        :param phone_code_hash: Telegram type: "string".
        :param next_type: Telegram type: "auth.CodeType".
        :param timeout: Telegram type: "int".

        Constructor for auth.SentCode: Instance of SentCode.
        """
        super().__init__()

        self.phone_registered = phone_registered
        self.type = type
        self.phone_code_hash = phone_code_hash
        self.next_type = next_type
        self.timeout = timeout

    def to_dict(self):
        return {
            'phone_registered': self.phone_registered,
            'type': None if self.type is None else self.type.to_dict(),
            'phone_code_hash': self.phone_code_hash,
            'next_type': None if self.next_type is None else self.next_type.to_dict(),
            'timeout': self.timeout,
        }

    def on_send(self, writer):
        writer.write_int(SentCode.constructor_id, signed=False)
        flags = 0
        flags |= (1 << 0) if self.phone_registered else 0
        flags |= (1 << 1) if self.next_type else 0
        flags |= (1 << 2) if self.timeout else 0
        writer.write_int(flags)

        self.type.on_send(writer)
        writer.tgwrite_string(self.phone_code_hash)
        if self.next_type:
            self.next_type.on_send(writer)

        if self.timeout:
            writer.write_int(self.timeout)

    @staticmethod
    def empty():
        return SentCode(None, None, None, None, None)

    def on_response(self, reader):
        flags = reader.read_int()

        if (flags & (1 << 0)) != 0:
            self.phone_registered = True

        self.type = reader.tgread_object()
        self.phone_code_hash = reader.tgread_string()
        if (flags & (1 << 1)) != 0:
            self.next_type = reader.tgread_object()

        if (flags & (1 << 2)) != 0:
            self.timeout = reader.read_int()

    def __repr__(self):
        return 'auth.sentCode#5e002502 flags:# phone_registered:flags.0?true type:auth.SentCodeType phone_code_hash:string next_type:flags.1?auth.CodeType timeout:flags.2?int = auth.SentCode'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class SentCodeTypeApp(TLObject):
    constructor_id = 0x3dbb5986
    subclass_of_id = 0xff5b158e

    def __init__(self, length):
        """
        :param length: Telegram type: "int".

        Constructor for auth.SentCodeType: Instance of either SentCodeTypeApp, SentCodeTypeSms, SentCodeTypeCall, SentCodeTypeFlashCall.
        """
        super().__init__()

        self.length = length

    def to_dict(self):
        return {
            'length': self.length,
        }

    def on_send(self, writer):
        writer.write_int(SentCodeTypeApp.constructor_id, signed=False)
        writer.write_int(self.length)

    @staticmethod
    def empty():
        return SentCodeTypeApp(None)

    def on_response(self, reader):
        self.length = reader.read_int()

    def __repr__(self):
        return 'auth.sentCodeTypeApp#3dbb5986 length:int = auth.SentCodeType'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class SentCodeTypeCall(TLObject):
    constructor_id = 0x5353e5a7
    subclass_of_id = 0xff5b158e

    def __init__(self, length):
        """
        :param length: Telegram type: "int".

        Constructor for auth.SentCodeType: Instance of either SentCodeTypeApp, SentCodeTypeSms, SentCodeTypeCall, SentCodeTypeFlashCall.
        """
        super().__init__()

        self.length = length

    def to_dict(self):
        return {
            'length': self.length,
        }

    def on_send(self, writer):
        writer.write_int(SentCodeTypeCall.constructor_id, signed=False)
        writer.write_int(self.length)

    @staticmethod
    def empty():
        return SentCodeTypeCall(None)

    def on_response(self, reader):
        self.length = reader.read_int()

    def __repr__(self):
        return 'auth.sentCodeTypeCall#5353e5a7 length:int = auth.SentCodeType'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class SentCodeTypeFlashCall(TLObject):
    constructor_id = 0xab03c6d9
    subclass_of_id = 0xff5b158e

    def __init__(self, pattern):
        """
        :param pattern: Telegram type: "string".

        Constructor for auth.SentCodeType: Instance of either SentCodeTypeApp, SentCodeTypeSms, SentCodeTypeCall, SentCodeTypeFlashCall.
        """
        super().__init__()

        self.pattern = pattern

    def to_dict(self):
        return {
            'pattern': self.pattern,
        }

    def on_send(self, writer):
        writer.write_int(SentCodeTypeFlashCall.constructor_id, signed=False)
        writer.tgwrite_string(self.pattern)

    @staticmethod
    def empty():
        return SentCodeTypeFlashCall(None)

    def on_response(self, reader):
        self.pattern = reader.tgread_string()

    def __repr__(self):
        return 'auth.sentCodeTypeFlashCall#ab03c6d9 pattern:string = auth.SentCodeType'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class SentCodeTypeSms(TLObject):
    constructor_id = 0xc000bba2
    subclass_of_id = 0xff5b158e

    def __init__(self, length):
        """
        :param length: Telegram type: "int".

        Constructor for auth.SentCodeType: Instance of either SentCodeTypeApp, SentCodeTypeSms, SentCodeTypeCall, SentCodeTypeFlashCall.
        """
        super().__init__()

        self.length = length

    def to_dict(self):
        return {
            'length': self.length,
        }

    def on_send(self, writer):
        writer.write_int(SentCodeTypeSms.constructor_id, signed=False)
        writer.write_int(self.length)

    @staticmethod
    def empty():
        return SentCodeTypeSms(None)

    def on_response(self, reader):
        self.length = reader.read_int()

    def __repr__(self):
        return 'auth.sentCodeTypeSms#c000bba2 length:int = auth.SentCodeType'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)




"""File generated by TLObjects' generator. All changes will be ERASED"""
from ...tl.tlobject import TLObject


class AdminLogResults(TLObject):
    constructor_id = 0xed8af74d
    subclass_of_id = 0x51f076bc

    def __init__(self, events, chats, users):
        """
        :param events: Telegram type: "ChannelAdminLogEvent". Must be a list.
        :param chats: Telegram type: "Chat". Must be a list.
        :param users: Telegram type: "User". Must be a list.

        Constructor for channels.AdminLogResults: Instance of AdminLogResults.
        """
        super().__init__()

        self.events = events
        self.chats = chats
        self.users = users

    def to_dict(self):
        return {
            'events': [] if self.events is None else [None if x is None else x.to_dict() for x in self.events],
            'chats': [] if self.chats is None else [None if x is None else x.to_dict() for x in self.chats],
            'users': [] if self.users is None else [None if x is None else x.to_dict() for x in self.users],
        }

    def on_send(self, writer):
        writer.write_int(AdminLogResults.constructor_id, signed=False)
        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.events))
        for _x in self.events:
            _x.on_send(writer)

        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.chats))
        for _x in self.chats:
            _x.on_send(writer)

        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.users))
        for _x in self.users:
            _x.on_send(writer)

    @staticmethod
    def empty():
        return AdminLogResults(None, None, None)

    def on_response(self, reader):
        reader.read_int()
        self.events = []
        _len = reader.read_int()
        for _ in range(_len):
            _x = reader.tgread_object()
            self.events.append(_x)

        reader.read_int()
        self.chats = []
        _len = reader.read_int()
        for _ in range(_len):
            _x = reader.tgread_object()
            self.chats.append(_x)

        reader.read_int()
        self.users = []
        _len = reader.read_int()
        for _ in range(_len):
            _x = reader.tgread_object()
            self.users.append(_x)

    def __repr__(self):
        return 'channels.adminLogResults#ed8af74d events:Vector<ChannelAdminLogEvent> chats:Vector<Chat> users:Vector<User> = channels.AdminLogResults'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class ChannelParticipant(TLObject):
    constructor_id = 0xd0d9b163
    subclass_of_id = 0x6658151a

    def __init__(self, participant, users):
        """
        :param participant: Telegram type: "ChannelParticipant".
        :param users: Telegram type: "User". Must be a list.

        Constructor for channels.ChannelParticipant: Instance of ChannelParticipant.
        """
        super().__init__()

        self.participant = participant
        self.users = users

    def to_dict(self):
        return {
            'participant': None if self.participant is None else self.participant.to_dict(),
            'users': [] if self.users is None else [None if x is None else x.to_dict() for x in self.users],
        }

    def on_send(self, writer):
        writer.write_int(ChannelParticipant.constructor_id, signed=False)
        self.participant.on_send(writer)
        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.users))
        for _x in self.users:
            _x.on_send(writer)

    @staticmethod
    def empty():
        return ChannelParticipant(None, None)

    def on_response(self, reader):
        self.participant = reader.tgread_object()
        reader.read_int()
        self.users = []
        _len = reader.read_int()
        for _ in range(_len):
            _x = reader.tgread_object()
            self.users.append(_x)

    def __repr__(self):
        return 'channels.channelParticipant#d0d9b163 participant:ChannelParticipant users:Vector<User> = channels.ChannelParticipant'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class ChannelParticipants(TLObject):
    constructor_id = 0xf56ee2a8
    subclass_of_id = 0xe60a6e64

    def __init__(self, count, participants, users):
        """
        :param count: Telegram type: "int".
        :param participants: Telegram type: "ChannelParticipant". Must be a list.
        :param users: Telegram type: "User". Must be a list.

        Constructor for channels.ChannelParticipants: Instance of ChannelParticipants.
        """
        super().__init__()

        self.count = count
        self.participants = participants
        self.users = users

    def to_dict(self):
        return {
            'count': self.count,
            'participants': [] if self.participants is None else [None if x is None else x.to_dict() for x in self.participants],
            'users': [] if self.users is None else [None if x is None else x.to_dict() for x in self.users],
        }

    def on_send(self, writer):
        writer.write_int(ChannelParticipants.constructor_id, signed=False)
        writer.write_int(self.count)
        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.participants))
        for _x in self.participants:
            _x.on_send(writer)

        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.users))
        for _x in self.users:
            _x.on_send(writer)

    @staticmethod
    def empty():
        return ChannelParticipants(None, None, None)

    def on_response(self, reader):
        self.count = reader.read_int()
        reader.read_int()
        self.participants = []
        _len = reader.read_int()
        for _ in range(_len):
            _x = reader.tgread_object()
            self.participants.append(_x)

        reader.read_int()
        self.users = []
        _len = reader.read_int()
        for _ in range(_len):
            _x = reader.tgread_object()
            self.users.append(_x)

    def __repr__(self):
        return 'channels.channelParticipants#f56ee2a8 count:int participants:Vector<ChannelParticipant> users:Vector<User> = channels.ChannelParticipants'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)




"""File generated by TLObjects' generator. All changes will be ERASED"""
from ...tl.tlobject import TLObject
from . import contest, auth, account, users, contacts, messages, updates, photos, upload, help, channels, bots, payments, stickers, phone, langpack


class DestroyAuthKeyRequest(TLObject):
    constructor_id = 0xd1435160
    subclass_of_id = 0x8291e68e

    def __init__(self):
        super().__init__()
        self.result = None
        self.content_related = True

    @staticmethod
    def to_dict():
        return {}

    def on_send(self, writer):
        writer.write_int(DestroyAuthKeyRequest.constructor_id, signed=False)

    @staticmethod
    def empty():
        return DestroyAuthKeyRequest()

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'destroy_auth_key#d1435160 = DestroyAuthKeyRes'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class DestroySessionRequest(TLObject):
    constructor_id = 0xe7512126
    subclass_of_id = 0xaf0ce7bd

    def __init__(self, session_id):
        """
        :param session_id: Telegram type: "long".

        :returns DestroySessionRes: Instance of either DestroySessionOk, DestroySessionNone.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.session_id = session_id

    def to_dict(self):
        return {
            'session_id': self.session_id,
        }

    def on_send(self, writer):
        writer.write_int(DestroySessionRequest.constructor_id, signed=False)
        writer.write_long(self.session_id)

    @staticmethod
    def empty():
        return DestroySessionRequest(None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'destroy_session#e7512126 session_id:long = DestroySessionRes'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class GetFutureSaltsRequest(TLObject):
    constructor_id = 0xb921bd04
    subclass_of_id = 0x1090f517

    def __init__(self, num):
        """
        :param num: Telegram type: "int".

        :returns FutureSalts: Instance of FutureSalts.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.num = num

    def to_dict(self):
        return {
            'num': self.num,
        }

    def on_send(self, writer):
        writer.write_int(GetFutureSaltsRequest.constructor_id, signed=False)
        writer.write_int(self.num)

    @staticmethod
    def empty():
        return GetFutureSaltsRequest(None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'get_future_salts#b921bd04 num:int = FutureSalts'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class InitConnectionRequest(TLObject):
    constructor_id = 0xc7481da6
    subclass_of_id = 0xb7b2364b

    def __init__(self, api_id, device_model, system_version, app_version, system_lang_code, lang_pack, lang_code, query):
        """
        :param api_id: Telegram type: "int".
        :param device_model: Telegram type: "string".
        :param system_version: Telegram type: "string".
        :param app_version: Telegram type: "string".
        :param system_lang_code: Telegram type: "string".
        :param lang_pack: Telegram type: "string".
        :param lang_code: Telegram type: "string".
        :param query: Telegram type: "X". Must be another TLObject request.

        :returns X: This type has no constructors.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.api_id = api_id
        self.device_model = device_model
        self.system_version = system_version
        self.app_version = app_version
        self.system_lang_code = system_lang_code
        self.lang_pack = lang_pack
        self.lang_code = lang_code
        self.query = query

    def to_dict(self):
        return {
            'api_id': self.api_id,
            'device_model': self.device_model,
            'system_version': self.system_version,
            'app_version': self.app_version,
            'system_lang_code': self.system_lang_code,
            'lang_pack': self.lang_pack,
            'lang_code': self.lang_code,
            'query': None if self.query is None else self.query.to_dict(),
        }

    def on_send(self, writer):
        writer.write_int(InitConnectionRequest.constructor_id, signed=False)
        writer.write_int(self.api_id)
        writer.tgwrite_string(self.device_model)
        writer.tgwrite_string(self.system_version)
        writer.tgwrite_string(self.app_version)
        writer.tgwrite_string(self.system_lang_code)
        writer.tgwrite_string(self.lang_pack)
        writer.tgwrite_string(self.lang_code)
        self.query.on_send(writer)

    @staticmethod
    def empty():
        return InitConnectionRequest(None, None, None, None, None, None, None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'initConnection#c7481da6 {X:Type} api_id:int device_model:string system_version:string app_version:string system_lang_code:string lang_pack:string lang_code:string query:!X = X'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class InvokeAfterMsgRequest(TLObject):
    constructor_id = 0xcb9f372d
    subclass_of_id = 0xb7b2364b

    def __init__(self, msg_id, query):
        """
        :param msg_id: Telegram type: "long".
        :param query: Telegram type: "X". Must be another TLObject request.

        :returns X: This type has no constructors.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.msg_id = msg_id
        self.query = query

    def to_dict(self):
        return {
            'msg_id': self.msg_id,
            'query': None if self.query is None else self.query.to_dict(),
        }

    def on_send(self, writer):
        writer.write_int(InvokeAfterMsgRequest.constructor_id, signed=False)
        writer.write_long(self.msg_id)
        self.query.on_send(writer)

    @staticmethod
    def empty():
        return InvokeAfterMsgRequest(None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'invokeAfterMsg#cb9f372d {X:Type} msg_id:long query:!X = X'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class InvokeAfterMsgsRequest(TLObject):
    constructor_id = 0x3dc4b4f0
    subclass_of_id = 0xb7b2364b

    def __init__(self, msg_ids, query):
        """
        :param msg_ids: Telegram type: "long". Must be a list.
        :param query: Telegram type: "X". Must be another TLObject request.

        :returns X: This type has no constructors.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.msg_ids = msg_ids
        self.query = query

    def to_dict(self):
        return {
            'msg_ids': [] if self.msg_ids is None else self.msg_ids[:],
            'query': None if self.query is None else self.query.to_dict(),
        }

    def on_send(self, writer):
        writer.write_int(InvokeAfterMsgsRequest.constructor_id, signed=False)
        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.msg_ids))
        for _x in self.msg_ids:
            writer.write_long(_x)

        self.query.on_send(writer)

    @staticmethod
    def empty():
        return InvokeAfterMsgsRequest(None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'invokeAfterMsgs#3dc4b4f0 {X:Type} msg_ids:Vector<long> query:!X = X'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class InvokeWithLayerRequest(TLObject):
    constructor_id = 0xda9b0d0d
    subclass_of_id = 0xb7b2364b

    def __init__(self, layer, query):
        """
        :param layer: Telegram type: "int".
        :param query: Telegram type: "X". Must be another TLObject request.

        :returns X: This type has no constructors.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.layer = layer
        self.query = query

    def to_dict(self):
        return {
            'layer': self.layer,
            'query': None if self.query is None else self.query.to_dict(),
        }

    def on_send(self, writer):
        writer.write_int(InvokeWithLayerRequest.constructor_id, signed=False)
        writer.write_int(self.layer)
        self.query.on_send(writer)

    @staticmethod
    def empty():
        return InvokeWithLayerRequest(None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'invokeWithLayer#da9b0d0d {X:Type} layer:int query:!X = X'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class InvokeWithoutUpdatesRequest(TLObject):
    constructor_id = 0xbf9459b7
    subclass_of_id = 0xb7b2364b

    def __init__(self, query):
        """
        :param query: Telegram type: "X". Must be another TLObject request.

        :returns X: This type has no constructors.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.query = query

    def to_dict(self):
        return {
            'query': None if self.query is None else self.query.to_dict(),
        }

    def on_send(self, writer):
        writer.write_int(InvokeWithoutUpdatesRequest.constructor_id, signed=False)
        self.query.on_send(writer)

    @staticmethod
    def empty():
        return InvokeWithoutUpdatesRequest(None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'invokeWithoutUpdates#bf9459b7 {X:Type} query:!X = X'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class PingRequest(TLObject):
    constructor_id = 0x7abe77ec
    subclass_of_id = 0x816aee71

    def __init__(self, ping_id):
        """
        :param ping_id: Telegram type: "long".

        :returns Pong: Instance of Pong.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.ping_id = ping_id

    def to_dict(self):
        return {
            'ping_id': self.ping_id,
        }

    def on_send(self, writer):
        writer.write_int(PingRequest.constructor_id, signed=False)
        writer.write_long(self.ping_id)

    @staticmethod
    def empty():
        return PingRequest(None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'ping#7abe77ec ping_id:long = Pong'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class PingDelayDisconnectRequest(TLObject):
    constructor_id = 0xf3427b8c
    subclass_of_id = 0x816aee71

    def __init__(self, ping_id, disconnect_delay):
        """
        :param ping_id: Telegram type: "long".
        :param disconnect_delay: Telegram type: "int".

        :returns Pong: Instance of Pong.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.ping_id = ping_id
        self.disconnect_delay = disconnect_delay

    def to_dict(self):
        return {
            'ping_id': self.ping_id,
            'disconnect_delay': self.disconnect_delay,
        }

    def on_send(self, writer):
        writer.write_int(PingDelayDisconnectRequest.constructor_id, signed=False)
        writer.write_long(self.ping_id)
        writer.write_int(self.disconnect_delay)

    @staticmethod
    def empty():
        return PingDelayDisconnectRequest(None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'ping_delay_disconnect#f3427b8c ping_id:long disconnect_delay:int = Pong'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class ReqDHParamsRequest(TLObject):
    constructor_id = 0xd712e4be
    subclass_of_id = 0xa6188d9e

    def __init__(self, nonce, server_nonce, p, q, public_key_fingerprint, encrypted_data):
        """
        :param nonce: Telegram type: "int128".
        :param server_nonce: Telegram type: "int128".
        :param p: Telegram type: "string".
        :param q: Telegram type: "string".
        :param public_key_fingerprint: Telegram type: "long".
        :param encrypted_data: Telegram type: "string".

        :returns Server_DH_Params: Instance of either ServerDHParamsFail, ServerDHParamsOk.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.nonce = nonce
        self.server_nonce = server_nonce
        self.p = p
        self.q = q
        self.public_key_fingerprint = public_key_fingerprint
        self.encrypted_data = encrypted_data

    def to_dict(self):
        return {
            'nonce': self.nonce,
            'server_nonce': self.server_nonce,
            'p': self.p,
            'q': self.q,
            'public_key_fingerprint': self.public_key_fingerprint,
            'encrypted_data': self.encrypted_data,
        }

    def on_send(self, writer):
        writer.write_int(ReqDHParamsRequest.constructor_id, signed=False)
        writer.write_large_int(self.nonce, bits=128)
        writer.write_large_int(self.server_nonce, bits=128)
        writer.tgwrite_string(self.p)
        writer.tgwrite_string(self.q)
        writer.write_long(self.public_key_fingerprint)
        writer.tgwrite_string(self.encrypted_data)

    @staticmethod
    def empty():
        return ReqDHParamsRequest(None, None, None, None, None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'req_DH_params#d712e4be nonce:int128 server_nonce:int128 p:string q:string public_key_fingerprint:long encrypted_data:string = Server_DH_Params'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class ReqPqRequest(TLObject):
    constructor_id = 0x60469778
    subclass_of_id = 0x786986b8

    def __init__(self, nonce):
        """
        :param nonce: Telegram type: "int128".

        :returns ResPQ: Instance of ResPQ.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.nonce = nonce

    def to_dict(self):
        return {
            'nonce': self.nonce,
        }

    def on_send(self, writer):
        writer.write_int(ReqPqRequest.constructor_id, signed=False)
        writer.write_large_int(self.nonce, bits=128)

    @staticmethod
    def empty():
        return ReqPqRequest(None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'req_pq#60469778 nonce:int128 = ResPQ'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class RpcDropAnswerRequest(TLObject):
    constructor_id = 0x58e4a740
    subclass_of_id = 0x4bca7570

    def __init__(self, req_msg_id):
        """
        :param req_msg_id: Telegram type: "long".

        :returns RpcDropAnswer: Instance of either RpcAnswerUnknown, RpcAnswerDroppedRunning, RpcAnswerDropped.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.req_msg_id = req_msg_id

    def to_dict(self):
        return {
            'req_msg_id': self.req_msg_id,
        }

    def on_send(self, writer):
        writer.write_int(RpcDropAnswerRequest.constructor_id, signed=False)
        writer.write_long(self.req_msg_id)

    @staticmethod
    def empty():
        return RpcDropAnswerRequest(None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'rpc_drop_answer#58e4a740 req_msg_id:long = RpcDropAnswer'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class SetClientDHParamsRequest(TLObject):
    constructor_id = 0xf5045f1f
    subclass_of_id = 0x55dd6cdb

    def __init__(self, nonce, server_nonce, encrypted_data):
        """
        :param nonce: Telegram type: "int128".
        :param server_nonce: Telegram type: "int128".
        :param encrypted_data: Telegram type: "string".

        :returns Set_client_DH_params_answer: Instance of either DhGenOk, DhGenRetry, DhGenFail.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.nonce = nonce
        self.server_nonce = server_nonce
        self.encrypted_data = encrypted_data

    def to_dict(self):
        return {
            'nonce': self.nonce,
            'server_nonce': self.server_nonce,
            'encrypted_data': self.encrypted_data,
        }

    def on_send(self, writer):
        writer.write_int(SetClientDHParamsRequest.constructor_id, signed=False)
        writer.write_large_int(self.nonce, bits=128)
        writer.write_large_int(self.server_nonce, bits=128)
        writer.tgwrite_string(self.encrypted_data)

    @staticmethod
    def empty():
        return SetClientDHParamsRequest(None, None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'set_client_DH_params#f5045f1f nonce:int128 server_nonce:int128 encrypted_data:string = Set_client_DH_params_answer'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)




from hashlib import sha256

from ..tl import Session
from ..tl.functions.upload import GetCdnFileRequest, ReuploadCdnFileRequest
from ..tl.types.upload import CdnFileReuploadNeeded, CdnFile
from ..crypto import AESModeCTR
from ..errors import CdnFileTamperedError


class CdnDecrypter:
    """Used when downloading a file results in a 'FileCdnRedirect' to
       both prepare the redirect, decrypt the file as it downloads, and
       ensure the file hasn't been tampered.
    """
    def __init__(self, cdn_client, file_token, cdn_aes, cdn_file_hashes):
        self.client = cdn_client
        self.file_token = file_token
        self.cdn_aes = cdn_aes
        self.cdn_file_hashes = cdn_file_hashes
        self.shaes = [sha256() for _ in range(len(cdn_file_hashes))]

    @staticmethod
    def prepare_decrypter(client, client_cls, cdn_redirect):
        """Prepares a CDN decrypter, returning (decrypter, file data).
           'client' should be the original TelegramBareClient that
           tried to download the file.

           'client_cls' should be the class of the TelegramBareClient.
        """
        # TODO Avoid the need for 'client_cls=TelegramBareClient'
        # https://core.telegram.org/cdn
        cdn_aes = AESModeCTR(
            key=cdn_redirect.encryption_key,
            # 12 first bytes of the IV..4 bytes of the offset (0, big endian)
            iv=cdn_redirect.encryption_iv[:12] + bytes(4)
        )

        # Create a new client on said CDN
        dc = client._get_dc(cdn_redirect.dc_id, cdn=True)
        session = Session(client.session)
        session.server_address = dc.ip_address
        session.port = dc.port
        cdn_client = client_cls(  # Avoid importing TelegramBareClient
            session, client.api_id, client.api_hash,
            timeout=client._timeout
        )
        # This will make use of the new RSA keys for this specific CDN.
        #
        # We assume that cdn_redirect.cdn_file_hashes are ordered by offset,
        # and that there will be enough of these to retrieve the whole file.
        cdn_file = cdn_client.connect(initial_query=GetCdnFileRequest(
            file_token=cdn_redirect.file_token,
            offset=cdn_redirect.cdn_file_hashes[0].offset,
            limit=cdn_redirect.cdn_file_hashes[0].limit
        ))

        # CDN client is ready, create the resulting CdnDecrypter
        decrypter = CdnDecrypter(
            cdn_client, cdn_redirect.file_token,
            cdn_aes, cdn_redirect.cdn_file_hashes
        )

        if isinstance(cdn_file, CdnFileReuploadNeeded):
            # We need to use the original client here
            client(ReuploadCdnFileRequest(
                file_token=cdn_redirect.file_token,
                request_token=cdn_file.request_token
            ))

            # We want to always return a valid upload.CdnFile
            cdn_file = decrypter.get_file()
        else:
            cdn_file.bytes = decrypter.cdn_aes.encrypt(cdn_file.bytes)
            cdn_hash = decrypter.cdn_file_hashes.pop(0)
            decrypter.check(cdn_file.bytes, cdn_hash)

        return decrypter, cdn_file

    def get_file(self):
        """Calls GetCdnFileRequest and decrypts its bytes.
           Also ensures that the file hasn't been tampered.
        """
        if self.cdn_file_hashes:
            cdn_hash = self.cdn_file_hashes.pop(0)
            cdn_file = self.client(GetCdnFileRequest(
                self.file_token, cdn_hash.offset, cdn_hash.limit
            ))
            cdn_file.bytes = self.cdn_aes.encrypt(cdn_file.bytes)
            self.check(cdn_file.bytes, cdn_hash)
        else:
            cdn_file = CdnFile(bytes(0))

        return cdn_file

    @staticmethod
    def check(data, cdn_hash):
        """Checks the integrity of the given data"""
        if sha256(data).digest() != cdn_hash.hash:
            raise CdnFileTamperedError()



"""File generated by TLObjects' generator. All changes will be ERASED"""
from ...tl.tlobject import TLObject


class BindTempAuthKeyRequest(TLObject):
    constructor_id = 0xcdd42a05
    subclass_of_id = 0xf5b399ac

    def __init__(self, perm_auth_key_id, nonce, expires_at, encrypted_message):
        """
        :param perm_auth_key_id: Telegram type: "long".
        :param nonce: Telegram type: "long".
        :param expires_at: Telegram type: "date".
        :param encrypted_message: Telegram type: "bytes".

        :returns Bool: This type has no constructors.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.perm_auth_key_id = perm_auth_key_id
        self.nonce = nonce
        self.expires_at = expires_at
        self.encrypted_message = encrypted_message

    def to_dict(self):
        return {
            'perm_auth_key_id': self.perm_auth_key_id,
            'nonce': self.nonce,
            'expires_at': self.expires_at,
            'encrypted_message': self.encrypted_message,
        }

    def on_send(self, writer):
        writer.write_int(BindTempAuthKeyRequest.constructor_id, signed=False)
        writer.write_long(self.perm_auth_key_id)
        writer.write_long(self.nonce)
        writer.tgwrite_date(self.expires_at)
        writer.tgwrite_bytes(self.encrypted_message)

    @staticmethod
    def empty():
        return BindTempAuthKeyRequest(None, None, None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'auth.bindTempAuthKey#cdd42a05 perm_auth_key_id:long nonce:long expires_at:int encrypted_message:bytes = Bool'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class CancelCodeRequest(TLObject):
    constructor_id = 0x1f040578
    subclass_of_id = 0xf5b399ac

    def __init__(self, phone_number, phone_code_hash):
        """
        :param phone_number: Telegram type: "string".
        :param phone_code_hash: Telegram type: "string".

        :returns Bool: This type has no constructors.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.phone_number = phone_number
        self.phone_code_hash = phone_code_hash

    def to_dict(self):
        return {
            'phone_number': self.phone_number,
            'phone_code_hash': self.phone_code_hash,
        }

    def on_send(self, writer):
        writer.write_int(CancelCodeRequest.constructor_id, signed=False)
        writer.tgwrite_string(self.phone_number)
        writer.tgwrite_string(self.phone_code_hash)

    @staticmethod
    def empty():
        return CancelCodeRequest(None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'auth.cancelCode#1f040578 phone_number:string phone_code_hash:string = Bool'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class CheckPasswordRequest(TLObject):
    constructor_id = 0xa63011e
    subclass_of_id = 0xb9e04e39

    def __init__(self, password_hash):
        """
        :param password_hash: Telegram type: "bytes".

        :returns auth.Authorization: Instance of Authorization.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.password_hash = password_hash

    def to_dict(self):
        return {
            'password_hash': self.password_hash,
        }

    def on_send(self, writer):
        writer.write_int(CheckPasswordRequest.constructor_id, signed=False)
        writer.tgwrite_bytes(self.password_hash)

    @staticmethod
    def empty():
        return CheckPasswordRequest(None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'auth.checkPassword#0a63011e password_hash:bytes = auth.Authorization'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class CheckPhoneRequest(TLObject):
    constructor_id = 0x6fe51dfb
    subclass_of_id = 0x99a3d765

    def __init__(self, phone_number):
        """
        :param phone_number: Telegram type: "string".

        :returns auth.CheckedPhone: Instance of CheckedPhone.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.phone_number = phone_number

    def to_dict(self):
        return {
            'phone_number': self.phone_number,
        }

    def on_send(self, writer):
        writer.write_int(CheckPhoneRequest.constructor_id, signed=False)
        writer.tgwrite_string(self.phone_number)

    @staticmethod
    def empty():
        return CheckPhoneRequest(None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'auth.checkPhone#6fe51dfb phone_number:string = auth.CheckedPhone'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class DropTempAuthKeysRequest(TLObject):
    constructor_id = 0x8e48a188
    subclass_of_id = 0xf5b399ac

    def __init__(self, except_auth_keys):
        """
        :param except_auth_keys: Telegram type: "long". Must be a list.

        :returns Bool: This type has no constructors.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.except_auth_keys = except_auth_keys

    def to_dict(self):
        return {
            'except_auth_keys': [] if self.except_auth_keys is None else self.except_auth_keys[:],
        }

    def on_send(self, writer):
        writer.write_int(DropTempAuthKeysRequest.constructor_id, signed=False)
        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.except_auth_keys))
        for _x in self.except_auth_keys:
            writer.write_long(_x)

    @staticmethod
    def empty():
        return DropTempAuthKeysRequest(None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'auth.dropTempAuthKeys#8e48a188 except_auth_keys:Vector<long> = Bool'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class ExportAuthorizationRequest(TLObject):
    constructor_id = 0xe5bfffcd
    subclass_of_id = 0x5fd1ec51

    def __init__(self, dc_id):
        """
        :param dc_id: Telegram type: "int".

        :returns auth.ExportedAuthorization: Instance of ExportedAuthorization.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.dc_id = dc_id

    def to_dict(self):
        return {
            'dc_id': self.dc_id,
        }

    def on_send(self, writer):
        writer.write_int(ExportAuthorizationRequest.constructor_id, signed=False)
        writer.write_int(self.dc_id)

    @staticmethod
    def empty():
        return ExportAuthorizationRequest(None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'auth.exportAuthorization#e5bfffcd dc_id:int = auth.ExportedAuthorization'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class ImportAuthorizationRequest(TLObject):
    constructor_id = 0xe3ef9613
    subclass_of_id = 0xb9e04e39

    def __init__(self, id, bytes):
        """
        :param id: Telegram type: "int".
        :param bytes: Telegram type: "bytes".

        :returns auth.Authorization: Instance of Authorization.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.id = id
        self.bytes = bytes

    def to_dict(self):
        return {
            'id': self.id,
            'bytes': self.bytes,
        }

    def on_send(self, writer):
        writer.write_int(ImportAuthorizationRequest.constructor_id, signed=False)
        writer.write_int(self.id)
        writer.tgwrite_bytes(self.bytes)

    @staticmethod
    def empty():
        return ImportAuthorizationRequest(None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'auth.importAuthorization#e3ef9613 id:int bytes:bytes = auth.Authorization'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class ImportBotAuthorizationRequest(TLObject):
    constructor_id = 0x67a3ff2c
    subclass_of_id = 0xb9e04e39

    def __init__(self, flags, api_id, api_hash, bot_auth_token):
        """
        :param flags: Telegram type: "int".
        :param api_id: Telegram type: "int".
        :param api_hash: Telegram type: "string".
        :param bot_auth_token: Telegram type: "string".

        :returns auth.Authorization: Instance of Authorization.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.flags = flags
        self.api_id = api_id
        self.api_hash = api_hash
        self.bot_auth_token = bot_auth_token

    def to_dict(self):
        return {
            'flags': self.flags,
            'api_id': self.api_id,
            'api_hash': self.api_hash,
            'bot_auth_token': self.bot_auth_token,
        }

    def on_send(self, writer):
        writer.write_int(ImportBotAuthorizationRequest.constructor_id, signed=False)
        writer.write_int(self.flags)
        writer.write_int(self.api_id)
        writer.tgwrite_string(self.api_hash)
        writer.tgwrite_string(self.bot_auth_token)

    @staticmethod
    def empty():
        return ImportBotAuthorizationRequest(None, None, None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'auth.importBotAuthorization#67a3ff2c flags:int api_id:int api_hash:string bot_auth_token:string = auth.Authorization'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class LogOutRequest(TLObject):
    constructor_id = 0x5717da40
    subclass_of_id = 0xf5b399ac

    def __init__(self):
        super().__init__()
        self.result = None
        self.content_related = True

    @staticmethod
    def to_dict():
        return {}

    def on_send(self, writer):
        writer.write_int(LogOutRequest.constructor_id, signed=False)

    @staticmethod
    def empty():
        return LogOutRequest()

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'auth.logOut#5717da40 = Bool'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class RecoverPasswordRequest(TLObject):
    constructor_id = 0x4ea56e92
    subclass_of_id = 0xb9e04e39

    def __init__(self, code):
        """
        :param code: Telegram type: "string".

        :returns auth.Authorization: Instance of Authorization.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.code = code

    def to_dict(self):
        return {
            'code': self.code,
        }

    def on_send(self, writer):
        writer.write_int(RecoverPasswordRequest.constructor_id, signed=False)
        writer.tgwrite_string(self.code)

    @staticmethod
    def empty():
        return RecoverPasswordRequest(None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'auth.recoverPassword#4ea56e92 code:string = auth.Authorization'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class RequestPasswordRecoveryRequest(TLObject):
    constructor_id = 0xd897bc66
    subclass_of_id = 0xfa72d43a

    def __init__(self):
        super().__init__()
        self.result = None
        self.content_related = True

    @staticmethod
    def to_dict():
        return {}

    def on_send(self, writer):
        writer.write_int(RequestPasswordRecoveryRequest.constructor_id, signed=False)

    @staticmethod
    def empty():
        return RequestPasswordRecoveryRequest()

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'auth.requestPasswordRecovery#d897bc66 = auth.PasswordRecovery'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class ResendCodeRequest(TLObject):
    constructor_id = 0x3ef1a9bf
    subclass_of_id = 0x6ce87081

    def __init__(self, phone_number, phone_code_hash):
        """
        :param phone_number: Telegram type: "string".
        :param phone_code_hash: Telegram type: "string".

        :returns auth.SentCode: Instance of SentCode.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.phone_number = phone_number
        self.phone_code_hash = phone_code_hash

    def to_dict(self):
        return {
            'phone_number': self.phone_number,
            'phone_code_hash': self.phone_code_hash,
        }

    def on_send(self, writer):
        writer.write_int(ResendCodeRequest.constructor_id, signed=False)
        writer.tgwrite_string(self.phone_number)
        writer.tgwrite_string(self.phone_code_hash)

    @staticmethod
    def empty():
        return ResendCodeRequest(None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'auth.resendCode#3ef1a9bf phone_number:string phone_code_hash:string = auth.SentCode'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class ResetAuthorizationsRequest(TLObject):
    constructor_id = 0x9fab0d1a
    subclass_of_id = 0xf5b399ac

    def __init__(self):
        super().__init__()
        self.result = None
        self.content_related = True

    @staticmethod
    def to_dict():
        return {}

    def on_send(self, writer):
        writer.write_int(ResetAuthorizationsRequest.constructor_id, signed=False)

    @staticmethod
    def empty():
        return ResetAuthorizationsRequest()

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'auth.resetAuthorizations#9fab0d1a = Bool'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class SendCodeRequest(TLObject):
    constructor_id = 0x86aef0ec
    subclass_of_id = 0x6ce87081

    def __init__(self, phone_number, api_id, api_hash, allow_flashcall=None, current_number=None):
        """
        :param allow_flashcall: Telegram type: "true".
        :param phone_number: Telegram type: "string".
        :param current_number: Telegram type: "Bool".
        :param api_id: Telegram type: "int".
        :param api_hash: Telegram type: "string".

        :returns auth.SentCode: Instance of SentCode.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.allow_flashcall = allow_flashcall
        self.phone_number = phone_number
        self.current_number = current_number
        self.api_id = api_id
        self.api_hash = api_hash

    def to_dict(self):
        return {
            'allow_flashcall': self.allow_flashcall,
            'phone_number': self.phone_number,
            'current_number': self.current_number,
            'api_id': self.api_id,
            'api_hash': self.api_hash,
        }

    def on_send(self, writer):
        writer.write_int(SendCodeRequest.constructor_id, signed=False)
        flags = 0
        flags |= (1 << 0) if self.allow_flashcall else 0
        flags |= (1 << 0) if self.current_number else 0
        writer.write_int(flags)

        writer.tgwrite_string(self.phone_number)
        if self.current_number:
            writer.tgwrite_bool(self.current_number)

        writer.write_int(self.api_id)
        writer.tgwrite_string(self.api_hash)

    @staticmethod
    def empty():
        return SendCodeRequest(None, None, None, None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'auth.sendCode#86aef0ec flags:# allow_flashcall:flags.0?true phone_number:string current_number:flags.0?Bool api_id:int api_hash:string = auth.SentCode'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class SendInvitesRequest(TLObject):
    constructor_id = 0x771c1d97
    subclass_of_id = 0xf5b399ac

    def __init__(self, phone_numbers, message):
        """
        :param phone_numbers: Telegram type: "string". Must be a list.
        :param message: Telegram type: "string".

        :returns Bool: This type has no constructors.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.phone_numbers = phone_numbers
        self.message = message

    def to_dict(self):
        return {
            'phone_numbers': [] if self.phone_numbers is None else self.phone_numbers[:],
            'message': self.message,
        }

    def on_send(self, writer):
        writer.write_int(SendInvitesRequest.constructor_id, signed=False)
        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.phone_numbers))
        for _x in self.phone_numbers:
            writer.tgwrite_string(_x)

        writer.tgwrite_string(self.message)

    @staticmethod
    def empty():
        return SendInvitesRequest(None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'auth.sendInvites#771c1d97 phone_numbers:Vector<string> message:string = Bool'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class SignInRequest(TLObject):
    constructor_id = 0xbcd51581
    subclass_of_id = 0xb9e04e39

    def __init__(self, phone_number, phone_code_hash, phone_code):
        """
        :param phone_number: Telegram type: "string".
        :param phone_code_hash: Telegram type: "string".
        :param phone_code: Telegram type: "string".

        :returns auth.Authorization: Instance of Authorization.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.phone_number = phone_number
        self.phone_code_hash = phone_code_hash
        self.phone_code = phone_code

    def to_dict(self):
        return {
            'phone_number': self.phone_number,
            'phone_code_hash': self.phone_code_hash,
            'phone_code': self.phone_code,
        }

    def on_send(self, writer):
        writer.write_int(SignInRequest.constructor_id, signed=False)
        writer.tgwrite_string(self.phone_number)
        writer.tgwrite_string(self.phone_code_hash)
        writer.tgwrite_string(self.phone_code)

    @staticmethod
    def empty():
        return SignInRequest(None, None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'auth.signIn#bcd51581 phone_number:string phone_code_hash:string phone_code:string = auth.Authorization'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class SignUpRequest(TLObject):
    constructor_id = 0x1b067634
    subclass_of_id = 0xb9e04e39

    def __init__(self, phone_number, phone_code_hash, phone_code, first_name, last_name):
        """
        :param phone_number: Telegram type: "string".
        :param phone_code_hash: Telegram type: "string".
        :param phone_code: Telegram type: "string".
        :param first_name: Telegram type: "string".
        :param last_name: Telegram type: "string".

        :returns auth.Authorization: Instance of Authorization.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.phone_number = phone_number
        self.phone_code_hash = phone_code_hash
        self.phone_code = phone_code
        self.first_name = first_name
        self.last_name = last_name

    def to_dict(self):
        return {
            'phone_number': self.phone_number,
            'phone_code_hash': self.phone_code_hash,
            'phone_code': self.phone_code,
            'first_name': self.first_name,
            'last_name': self.last_name,
        }

    def on_send(self, writer):
        writer.write_int(SignUpRequest.constructor_id, signed=False)
        writer.tgwrite_string(self.phone_number)
        writer.tgwrite_string(self.phone_code_hash)
        writer.tgwrite_string(self.phone_code)
        writer.tgwrite_string(self.first_name)
        writer.tgwrite_string(self.last_name)

    @staticmethod
    def empty():
        return SignUpRequest(None, None, None, None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'auth.signUp#1b067634 phone_number:string phone_code_hash:string phone_code:string first_name:string last_name:string = auth.Authorization'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)

